/*
 * Copyright (c) 2003, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.binary;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;
import java.nio.charset.CodingErrorAction;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import nu.xom.Attribute;
import nu.xom.Comment;
import nu.xom.DocType;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.ProcessingInstruction;
import nu.xom.Text;

/**
 * Serializes (encodes) and deserializes (decodes) XOM XML documents to and from
 * an efficient and compact custom binary XML data format (termed <i>bnux </i>
 * format), without loss or change of any information. Serialization and
 * deserialization is much faster than with the standard textual XML format, and
 * the resulting binary data is more compressed than textual XML.
 * 
 * <h4>Applicability</h4>
 * 
 * The overall goal of the <i>bnux algorithm</i> is to maximize serialization and
 * deserialization (parsing) performance without requiring any schema
 * description. Serialization and deserialization speed are roughly balanced
 * against each other; neither side is particularly favoured over the other.
 * Another benefitial effect of the algorithm is that a considerable degree of
 * XML data redundancy is eliminated, but compression is more a welcome
 * side-effect than a primary goal in itself. The algorithm is primarily
 * intended for tightly coupled high-performance systems exchanging large
 * volumes of XML data over networks, as well as for compact main memory caches
 * and for <i>short-term </i> storage as BLOBs in backend databases or files 
 * (e.g. "session" data with limited duration). In the case of BLOB storage,
 * selecting matching BLOBs can be sped up by maintaining a simple metaindex 
 * side table for the most frequent access patterns.
 * See the <a href="#performance">performance results</a> below.
 * <p>
 * While the Java API is considered stable, the bnux data format should be
 * considered a black box: Its internals are under-documented and may change
 * without notice from release to release in backwards-incompatible manners. It
 * is unlikely that support for reading data written with older Nux versions
 * will ever be available. bnux is an exchange format but not an
 * interoperability format.
 * <p>
 * This approach is expressly <b>not intended </b>as a replacement for standard
 * textual XML in loosely coupled systems where maximum long-term
 * interoperability is the overarching concern. It is also expressly <b>not
 * intended </b>for long-term data storage. If you store data in bnux format
 * there's every chance you won't be able to read it back a year or two from
 * now, or even earlier. Finally, it is probably unwise to use this class if
 * your application's performance requirements are not particularly stringent,
 * or profiling indicates that the bottleneck is not related to XML
 * serialization/deserialization anyway.
 * <p>
 * The bnux serialization algorithm is a three-pass batch algorithm, hence
 * buffer-oriented, not stream-oriented. It has a throughput profile with short
 * critical paths, rather than a low latency profile with long critical paths,
 * rendering it ideal for large numbers of small to medium-sized XML documents,
 * and impractical for individual documents that do not fit into main memory.
 * The bnux deserialization algorithm is a single pass algorithm, and could in
 * theory be streamed through a {@link nu.xom.NodeFactory}, but the current
 * implementation does not do so.
 * 
 * <h4>Faithfully Preversing XML</h4>
 * 
 * Any and all arbitrary XOM XML documents are supported, and no schema is
 * required. A XOM document that is serialized and subsequently deserialized by
 * this class is <i>exactly the same </i> as the original document, preserving
 * "as is" all names and data for elements, namespaces, additional namespace
 * declarations, attributes, texts, document type, comments, processing
 * instructions, whitespace, Unicode characters, etc. As a result, the W3C XML
 * Infoset and the W3C Canonical XML representation is guaranteed to be
 * preserved. In particular there always holds:
 * 
 * <pre>
 * java.util.Arrays.equals(XOMUtil.toCanonicalXML(doc), XOMUtil
 * 		.toCanonicalXML(deserialize(serialize(doc))));
 * </pre>
 * 
 * <h4>Optional ZLIB Compression</h4>
 * 
 * The bnux algorithm considerably compresses XML data with little CPU
 * consumption, by its very design. However, bnux also has an option to further
 * compress/decompress its output/input with the ZLIB compression algorithm.
 * ZLIB is based on Huffman coding and also used by the popular
 * <code>gzip</code> (e.g. {@link java.util.zip.Deflater}). ZLIB compression
 * is rather CPU intensive, but it typically yields strong compression factors,
 * in particular for documents containing mostly narrative text (e.g. the
 * bible). For example, strong compression may be desirable over low-bandwith
 * networks or when bnux data is known to be accessed rather infrequently. On
 * the other hand, ZLIB compression probably kills performance in the presence
 * of high-bandwidth networks such as ESnet, Internet2/Abilene or 10 Gigabit Ethernet/InfiniBand LANs, 
 * even with high-end CPUs. CPU drain is also a scalability problem in the
 * presence of large amounts of concurrent connections. An option ranging from 0
 * (no ZLIB compression; best performance) to 1 (little ZLIB compression;
 * reduced performance) to 9 (strongest ZLIB compression; worst performance)
 * allows one to configure the CPU/memory consumption trade-off.
 * 
 * <h4>Reliability</h4>
 * 
 * This class has been successfully tested against many thousands of extremely
 * weird and unique test documents, including the W3C XML conformance test
 * suite, and no bugs are known.
 * <p>
 * Serialization employs no error checking at all, since malformed XOM input documents
 * are impossible to produce given XOM's design: XOM strictly enforces wellformedness anyway. 
 * Deserialization employs some limited error checking, throwing exceptions for
 * any improper API usage, non-bnux input data, data format version mismatch, or
 * general binary data corruption. Beyond this, deserialization relies on XOM's hard-wired 
 * wellformedness checks, just like serialization does.
 * Barring one of the above catastrophic situations, the bnux algorithm will always correctly and faithfully
 * reconstruct the exact same well-formed XOM document.
 * <p>
 * The implementation has one notable limitation, which is deemed irrelevant for
 * all practical purposes: XML data such as texts, element names, attribute
 * values, etc. MUST NOT contain a NIL character (char NIL = (char) 0x00). An
 * attempt to serialize a document containing a NIL character will be rejected
 * with an {@link IllegalArgumentException}. This limitation is not
 * conceptually inherent, it merely allows for more efficient implementation.
 * Any other arbitrary characters are fine, including Unicode surrogates.
 * 
 * <h4>Example Usage:</h4>
 * 
 * <pre>
 * // parse standard textual XML, convert to binary format, round-trip it and compare results
 * Document doc = new Builder().build(new File(&quot;/tmp/test.xml&quot;));
 * BinaryXMLCodec codec = new BinaryXMLCodec();
 * byte[] bnuxDoc = codec.serialize(doc, 0);
 * Document doc2 = codec.deserialize(bnuxDoc);
 * boolean isEqual = java.util.Arrays.equals(XOMUtil.toCanonicalXML(doc), XOMUtil
 * 		.toCanonicalXML(doc2));
 * System.out.println(&quot;isEqual = &quot; + isEqual);
 * System.out.println(doc2.toXML());
 * 
 * // write binary XML document to file
 * OutputStream out = new FileOutputStream(&quot;/tmp/test.xml.bnux&quot;);
 * out.write(bnuxDoc);
 * out.close();
 * 
 * // read binary XML document from file
 * bnuxDoc = XOMUtil.toByteArray(new FileInputStream(&quot;/tmp/test.xml.bnux&quot;));
 * Document doc3 = codec.deserialize(bnuxDoc);
 * System.out.println(doc3.toXML());
 * </pre>
 * 
 * <a name="performance"/>
 * <h4>Performance</h4>
 * 
 * This class has been carefully profiled and optimized. Preliminary performance
 * results over a wide range of real-world documents are given below.
 * <p>
 * Contrasting bnux BinaryXMLCodec with the XOM Builder, Serializer and toXML():
 * <ul>
 * <li>Serialization speedup: 15-50 MB/s vs. 4-5 MB/s</li>
 * <li>Deserialization speedup: 30-101 MB/s vs. 2-11 MB/s</li>
 * <li>Data compression factor: 1.2 - 4</li>
 * </ul>
 * For meaningful comparison, MB/s and compression factors are always given in
 * relation to the original standard textual XML file size.
 * <ul>
 * <li>Benchmark test data: A wide variety of small to large XML documents is
 * used, including SOAP documents heavily using namespaces ( <a target="_blank"
 * href="http://xbis.sourceforge.net/">XBIS </a>), simple XML formatted web
 * server logs using no namespaces, RDF documents with lots of attributes and
 * namespaces, the periodic table, documents consisting of large narrative text ( <a target="_blank"
 * href="http://www.oasis-open.org/cover/bosakShakespeare200.html">Shakespeare
 * </a>), publication citations ( <a target="_blank"
 * href="http://dblp.uni-trier.de/xml/">DBLP </a>), music title databases ( <a
 * target="_blank" href="http://www.freedb.org/">FreeDB </a>), Japanese
 * documents (XML conformance test suite), SVG image files, etc.</li>
 * 
 * <li>Benchmark configuration: no ZLIB compression, xom-1.0 final, non-validating
 * XOM Builder using xerces-2.6.2, no DTD or schema, SUN JDK 1.5.0 server VM,
 * commodity PC 2004, Dual Pentium 4, 2.8 GHz, Redhat 9</li>
 * </ul>
 * 
 * Example Interpretation:
 * <ul>
 * <li>Small documents: Results translate, for example, to ping-pong
 * round-tripping of typical 500 byte SOAP request/response message documents at
 * a rate of 25000 msg/s, compared to 2500 msg/s with XOM (excluding network
 * latency). More pronounced, 500 (150) byte documents with few namespaces translate to 
 * 35000 (120000) msg/s, compared to 3500 (5000) msg/s with XOM.
 * Consequently, XML serialization and deserialization are probably
 * nomore your application's bottleneck, leaving, say, 95% CPU headroom free for 
 * other application modules.</li>
 * 
 * <li>Large documents: When having a main-memory cache of several thousand 1
 * MB documents, each containing highly structured complex data, one can
 * deserialize, XQuery and serve from the cache at a rate of 50 documents/s,
 * compared to 5 documents/s with XOM.</li>
 * 
 * </ul>
 * Note that in contrast to other algorithms, bnux includes XOM tree building
 * and walking, hence measures delivering data to and from actual XML
 * applications, rather than merely to and from a low-level SAX event stream 
 * (which is considerably cheaper and deemed less useful).
 * <p>
 * The deserialization speedup is further multiplied when DTDs or schema
 * validation is used while parsing standard textual XML.
 * <p>
 * This class heavily relies on advanced Java compiler optimizations, which take
 * considerable time to warm up. Hence, for comparative benchmarks, use a
 * server-class VM and make sure to repeat runs for at least 30 seconds.
 * <p>
 * Further, you will probably want to eliminate drastic XOM hotspots by applying
 * two small backwards-compatible <a target="_blank"
 * href="http://dsd.lbl.gov/~hoschek/xom-patches/">patches </a> to the XOM
 * source code base. These patches 1) allow to selectively disable the
 * {@link nu.xom.Verifier}sanity checks for PCDATA via a System
 * property, and 2) maintain an internal String instead of an UTF-8 encoded
 * byte array in {@link nu.xom.Text}, which eliminates the expensive character
 * conversions implied for each access to a Text object. This increases
 * performance at the expense of memory footprint. The measurements above report
 * numbers using these patches, both for xom and bnux. If you're curious about
 * the whereabouts of bottlenecks, run java with the non-perturbing '-server
 * -agentlib:hprof=cpu=samples,depth=10' flags, then study the trace log and
 * correlate its hotspot trailer with its call stack headers (see <a
 * target="_blank"
 * href="http://java.sun.com/developer/technicalArticles/Programming/HPROF.html">
 * hprof tracing </a>).
 * <p>
 * Use class {@link nux.xom.tests.BinaryXMLTest}to reproduce results, verify
 * correctness or to evaluate performance for your own datasets.
 * 
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.78 $, $Date: 2005/02/12 21:26:12 $
 */
public class BinaryXMLCodec {

	// for deserialization: unique names from deserialized namepool
	private String[] nameList;

	// for deserialization and serialization: to convert UTF-8 to java string
	private CharBuffer tmpCharBuffer; // this is ALWAYS a HeapCharBuffer
	
	// for serialization:
	// holds unique strings found in document (qnames, texts, uris, etc.)
	private HashMap namePool; 
	
	// for serialization:
	// byte-level node type tokens in XML document order; also length of next index(es)
	private ArrayByteList nodeTokens;
	
	// for serialization: indexes into namepool/entries
	private ArrayIntList indexData; 
	
	// for serialization: output buffer
	private ArrayByteList output; // multi-byte integers are ALWAYS in big-endian

	// ZLIB
	private Inflater inflater; // ZLIB decompressor
	private Deflater deflater; // ZLIB compressor
	private int deflationLevel = -1; // initialize to "undefined"
	
	private static final Charset UTF8 = Charset.forName("UTF-8"); // construct once and for all
	
	// for serialization:
	private final CharsetDecoder utf8Decoder = UTF8.newDecoder()
		.onMalformedInput(CodingErrorAction.REPORT)
		.onUnmappableCharacter(CodingErrorAction.REPORT);

	// for deserialization:
	private final CharsetEncoder utf8Encoder = UTF8.newEncoder()
		.onMalformedInput(CodingErrorAction.REPLACE)
		.onUnmappableCharacter(CodingErrorAction.REPLACE);

	// for deserialization: avoids reverification of PCDATA
	private Text[] textCache;
	
	// for deserialization: 
	// avoids reverification of qname and URI, as well as indexOf() calls, and saves string memory
	private final ReadAttributeMicroCache readAttributeCache = new ReadAttributeMicroCache(); 
	
	// for serialization: used for string-interning of attribute qnames, speeding up namepool HashMap
	private final WriteAttributeMicroCache writeAttributeCache = new WriteAttributeMicroCache();
	
	// for deserialization:
	// avoids reverification of qname and URI, as well as indexOf() calls, and saves string memory
	private final ReadElementMicroCache readElementCache = new ReadElementMicroCache(); 

	// low 4 bits for node token types
	// high 4 bits typically hold number of bytes of the next two indexes
	private static final int TEXT = 0;
	private static final int FREQUENT_TEXT = 1;
	private static final int ATTRIBUTE = 2;
	private static final int ATTRIBUTE_WITH_NULL_NAMESPACE = 3;
	private static final int START_ELEMENT = 4;
	private static final int START_ELEMENT_WITH_NULL_NAMESPACE = 5;
	private static final int FINISH_ELEMENT = 6;
	private static final int COMMENT = 7;
	private static final int NAMESPACE_DECLARATION = 8;
	private static final int PROCESSING_INSTRUCTION = 9;
	private static final int DOCTYPE = 10;
	private static final int START_DOCUMENT = 11;
	private static final int FINISH_DOCUMENT = 12;
	
	// We exploit the fact that the UTF-8 spec guarantees that a zero byte is never 
	// introduced as a side-effect of UTF-8 encoding !
	private static final char NIL = (char) 0x00;
	
	private static final int BNUX_MAGIC_NUMBER = 11517829; // for sanity checks (first byte=0)
	private static final byte VERSION = 4; // version of bnux data format
	private static final int HEADER_SIZE = 4 + 1 + 4 + 4 + 4; // bnux header in bytes
	
	private static final boolean DEBUG = false;
		
	/**
	 * Constructs an instance of this class. An instance can be reused serially,
	 * but is not thread-safe, just like a {@link nu.xom.Builder}.
	 */
	public BinaryXMLCodec() {
	}
	
	/**
	 * Returns the XOM document obtained by deserializing the given binary XML
	 * document (<i>bnux</i> document).
	 * <p>
	 * If the document is in ZLIB compressed bnux format, it will be auto-detected
	 * and decompressed before applying deserialization.
	 * 
	 * @param bnuxDocument
	 *            the bnux document to deserialize.
	 * @return the new XOM document obtained from deserialization.
	 * @throws BinaryParsingException
	 *             if the bnux document is unreadable or corrupt for some reason
	 */
	public Document deserialize(byte[] bnuxDocument) throws BinaryParsingException {
		if (bnuxDocument == null) 
			throw new IllegalArgumentException("bnuxDocument must not be null");
		try {
			if (bnuxDocument.length < 2) 
				throw new BinaryParsingException("Missing bnux header");
			
			ByteBuffer buf;
			if ((bnuxDocument[0] & 0x0F) != Deflater.DEFLATED) { 
				buf = ByteBuffer.wrap(bnuxDocument); // it's not ZLIB compressed
			}
			else { // decompress with ZLIB
				buf = doInflate(bnuxDocument);
			}
	
			if (buf.remaining() < HEADER_SIZE) 
				throw new BinaryParsingException("Missing remaining bnux header");
		
			int magic = buf.getInt();
			if (magic != BNUX_MAGIC_NUMBER) throw new BinaryParsingException(
				"Bnux magic number mismatch: " + magic + ", must be: " + BNUX_MAGIC_NUMBER);
			byte version = buf.get();
			if (version != VERSION) throw new BinaryParsingException(
				"Bnux data format version mismatch: " + version + ", must be: " + VERSION);
			int namePoolSize = buf.getInt();
			if (namePoolSize < 0) 
				throw new BinaryParsingException("Negative name table size");
			int decodedSize = buf.getInt();
			if (decodedSize < 0) 
				throw new BinaryParsingException("Negative decodedSize");
			int encodedSize = buf.getInt();
			if (encodedSize < 0) 
				throw new BinaryParsingException("Negative encodedSize");
			if (buf.remaining() < encodedSize) 
				throw new BinaryParsingException("Missing bytes to decode");
			
			this.nameList = new String[namePoolSize];
			if (decodedSize == encodedSize)
				decodeASCIINames(buf, this.nameList, encodedSize); // safe trick, much faster!
			else 
				decodeUTFNames(buf, encodedSize, this.nameList, decodedSize);
			
			// reset caches in preparation for readDocument(buf)
			this.textCache = new Text[Math.min(256, namePoolSize)];
			this.readElementCache.reset();
			this.readAttributeCache.reset();
			
			// parse node token data and packed indexes, building the XOM tree
			return readDocument(buf);
		} catch (RuntimeException e) {
			throw new BinaryParsingException(e);
		} finally {
			this.nameList = null; // help gc
			this.textCache = null; // help gc
			if (DEBUG) System.err.println(this.readElementCache.toString());
		}
	}
	
	private void decodeASCIINames(ByteBuffer buf, String[] names, int encodedSize) {
		// efficient batch decoding of the namepool:
		int len = names.length;
		byte terminator = (byte) NIL;
		byte[] chars = buf.array();
		int off = buf.arrayOffset() + buf.position();
		
		for (int i = 0; i < len; i++) {
			int start = off;
			while (chars[off++] != terminator) ; 
			names[i] = new String(chars, 0, start, off-start-1);
		}
		buf.position(buf.position() + encodedSize);
	}
	
	private void decodeUTFNames(ByteBuffer buf, int encodedSize, String[] names, int decodedSize) throws BinaryParsingException {
		// efficient batch decoding of the namepool:
		if (tmpCharBuffer == null || tmpCharBuffer.capacity() < decodedSize) {
			tmpCharBuffer = CharBuffer.allocate(decodedSize);
		}
		tmpCharBuffer.clear();
		tmpCharBuffer.limit(decodedSize);
		int currLimit = buf.limit();
		buf.limit(buf.position() + encodedSize); // just read the namepool
		utf8Decoder.reset();
		
		CoderResult cr = utf8Decoder.decode(buf, tmpCharBuffer, true);
		
		if (cr.isError() || cr.isOverflow()) {
			try {
				cr.throwException();
			} catch (CharacterCodingException e) {
				throw new BinaryParsingException(e);
			}
		}
		utf8Decoder.flush(tmpCharBuffer);
		buf.limit(currLimit);
		decodeAndCopyNames(tmpCharBuffer.array(), names);
	}
	
	/** decode UTF-8 encoded namepool entries into namepool strings */
	private void decodeAndCopyNames(char[] chars, String[] names) {
		// scan and convert zero delimited strings
		int off = 0;
		int len = names.length;
		char terminator = NIL;
		
		for (int i = 0; i < len; i++) {
			int start = off;
			while (chars[off++] != terminator) ; 
			names[i] = new String(chars, start, off-start-1);
		}
	}
	
	/**
	 * Returns the binary XML document ( <i>bnux </i> document) obtained by
	 * serializing the given XOM document.
	 * <p>
	 * An optional zlib compression level ranging from 0 (no ZLIB compression;
	 * best performance) to 1 (little ZLIB compression; reduced performance) to
	 * 9 (strongest ZLIB compression; worst performance) allows one to configure
	 * the CPU/memory consumption trade-off.
	 * <p>
	 * Unless there is a good reason to the contrary, you should always use
	 * level 0: the bnux algorithm typically already precompresses considerably.
	 * 
	 * @param document
	 *            the XOM document to serialize
	 * @param zlibCompressionLevel
	 *            a number in the range 0..9
	 * @return the bnux document obtained from serialization.
	 * @throws IllegalArgumentException
	 *             if the compression level is out of range.
	 * @throws IllegalArgumentException
	 *             if the input document contains a NIL character (char NIL =
	 *             (char) 0x00).
	 */
	public byte[] serialize(Document document, int zlibCompressionLevel) {
		if (document == null) 
			throw new IllegalArgumentException("XOM document must not be null");
		if (zlibCompressionLevel < 0 || zlibCompressionLevel > 9)
			throw new IllegalArgumentException("Compression level must be 0..9");
		
		namePool = new HashMap();
		if (nodeTokens == null) nodeTokens = new ArrayByteList();
		nodeTokens.clear();
		if (indexData == null) indexData = new ArrayIntList(); 
		indexData.clear();
		writeAttributeCache.reset();

		writeDocument(document); // produces nodeTokens, indexData, namePool
		
		// retrieve entries of namepool
		Map.Entry[] entries = new Map.Entry[namePool.size()];
		int decodedSize = getPoolAsEntries(namePool, entries);
		namePool = null; // prevents next hashmap iteration degradation: iter is O(map capacity)
		
		// reorder entries and update indexData accordingly
		if (entries.length > 256) { 
			// if <= 256 there's no need to sort indexes - all fit into one byte anyway
			packSort(entries, indexData); 
		}

		// add header to output
		if (output == null) {
			output = new ArrayByteList(HEADER_SIZE);
		} else {
			output.clear();
			output.ensureCapacity(HEADER_SIZE);
		}
		ByteBuffer outputBuf = ByteBuffer.wrap(output.asArray(), 0, HEADER_SIZE);
		outputBuf.putInt(BNUX_MAGIC_NUMBER);
		outputBuf.put(VERSION);		
		outputBuf.putInt(entries.length);
		outputBuf.putInt(decodedSize);
		
		// add name pool to output		
		ByteBuffer names = encodeNames(entries, decodedSize);
		int encodedSize = names.remaining();
		outputBuf.putInt(encodedSize);
		output.setSize(outputBuf.position());
		
		// precompute memory size needed for worst case
		int maxLen;
		if (entries.length <= 256) maxLen = 1;
		else if  (entries.length <= 65536) maxLen = 2;
		else maxLen = 4;
		maxLen = encodedSize + nodeTokens.size() + maxLen*indexData.size();
		maxLen += (4 + 4 + 4 + 4); // DocType
//		System.out.println("entries.length="+entries.length);
//		System.out.println("entries[0..255]="+ Arrays.asList(entries).subList(0, 256));
//		System.out.println("\n\nentries[255..x]="+ Arrays.asList(entries).subList(256, entries.length));
//		System.out.println("entries="+ Arrays.asList(entries));
//		System.err.println("indexData.size()="+indexData.size());
//		System.err.println("encodedSize="+encodedSize);
		entries = null; // help gc

		output.ensureCapacity(output.size() + maxLen); // so outputbuf won't overflow
		outputBuf = ByteBuffer.wrap(output.asArray(), output.size(), maxLen);
		outputBuf.put(names);
		names = null; // help gc
		
		// add node tokens and packed name indexes to output
//		int pos = outputBuf.position();
		packNodeTokens(nodeTokens.asArray(), indexData.asArray(), outputBuf);		
		output.setSize(outputBuf.position());
//		System.err.println("tokensWithIndexes = "+ (outputBuf.position()-pos));
//		System.err.println("ratio of names / tokensWithIndexes = "+ (encodedSize * 1.0f / (outputBuf.position()-pos)));
				
		if (zlibCompressionLevel == 0) 
			return output.toArray();
		else
			return doDeflate(output, zlibCompressionLevel);
	}

	/** copy references to map entries into array */
	private int getPoolAsEntries(HashMap namePool, Map.Entry[] entries) {
		Iterator iter = namePool.entrySet().iterator();
		int decodedSize = 0;
		for (int i=entries.length; --i >= 0; ) {
			Map.Entry entry = (Map.Entry) iter.next();
			int j = ((StringInfo) entry.getValue()).index;
			entries[j] = entry;
			decodedSize += ((String) entry.getKey()).length() + 1;
		}
		return decodedSize; // length of all strings + string terminator
	}
	
	/** converts strings of namepool into UTF-8 bytes */
	private ByteBuffer encodeNames(Map.Entry[] entries, int decodedSize) {
		// set up char buffer array
		if (tmpCharBuffer == null || tmpCharBuffer.capacity() < decodedSize) {
			tmpCharBuffer = CharBuffer.allocate(decodedSize);
		} 
		else {
			tmpCharBuffer.clear();
			tmpCharBuffer.limit(decodedSize);
		}

		// copy entries into charbuffer array
		encodeAndCopyNames(entries, tmpCharBuffer.array());
		
		// now encode large string batch with UTF-8 in a single tight loop (!)
		// our encoding/decoding implementation does not exhibit JDK's 16 MB bug
		// http://developer.java.sun.com/developer/bugParade/bugs/4949631.html
		try {
			return utf8Encoder.encode(tmpCharBuffer);
		} catch (CharacterCodingException e) {
			throw new Error(e); // can never happen
		}
	}
	
	/** converts strings of namepool into array */
	private void encodeAndCopyNames(Map.Entry[] entries, char[] chars) {
		// concatenate all names into a single large string for efficient UTF batch encoding
		// use zero terminated strings instead of length prefixed strings for compactness
		char terminator = NIL;
		for (int i=0, size = 0, count = entries.length; i < count; i++) {
			String str = (String) entries[i].getKey();
			int len = str.length();
			str.getChars(0, len, chars, size);
			size += len;			
			chars[size++] = terminator;
		}
	}
	
	/** sort entries by name frequency (# of occurances) and update indexdata accordingly 
	 * this allows to compress frequent 4 byte indexes into a single byte 
	 * FAQ: this sort is *not* the bottleneck
	 */
	private void packSort(Map.Entry[] entries, ArrayIntList indexData) {
		Arrays.sort(entries,
			new Comparator() {
				public final int compare(Object a, Object b) { 
					Map.Entry e1 = (Map.Entry) a;
					Map.Entry e2 = (Map.Entry) b;
					int f1 = ((StringInfo) e1.getValue()).frequency;
					int f2 = ((StringInfo) e2.getValue()).frequency;
					return f2 - f1;
					//return f1 > f2 ? -1 : (f1 == f2 ? 0 : 1);
				}
			}
		);

		// reorder indexData with sorted indexes
		// since sort has moved entries[indexData[k]] to entries[i]
		int[] indexes = new int[entries.length];
		for (int i=entries.length; --i >= 0; ) {
			int k = ((StringInfo) entries[i].getValue()).index;
			indexes[k] = i;
		}
		
		int[] ix = indexData.asArray();
		for (int i=indexData.size(); --i >= 0; ) {
			ix[i] = indexes[ix[i]];
		}
		// post-condition: entries[indexData[k]] corresponds to entries[i]
	}	
	
	/** copy nodeTokens and stitch in packed indexes referring to names in the namepool */
	private void packNodeTokens(byte[] tokens, int[] indexes, ByteBuffer dest) {
		int i = 0;
		int j = 0;
		for (;;) {
			int pos = dest.position();
			int type = tokens[j++];
			dest.put((byte)type);
			
			switch (type & 0x0F) {			
				case TEXT: {
					packOneIndex(dest, pos, indexes, i++, type);
					break;
				}
				case FREQUENT_TEXT: {
					break;
				}
				case ATTRIBUTE: {
					packTwoIndexes(dest, pos, indexes, i, type);
					i += 2;
					dest.put(tokens[j++]);
					break;
				}
				case ATTRIBUTE_WITH_NULL_NAMESPACE: {
					packTwoIndexes(dest, pos, indexes, i, type);
					i += 2;
					dest.put(tokens[j++]);
					break;
				}
				case START_ELEMENT: {
					packTwoIndexes(dest, pos, indexes, i, type);
					i += 2;
					break;
				}
				case START_ELEMENT_WITH_NULL_NAMESPACE: {
					packOneIndex(dest, pos, indexes, i++, type);
					break;
				}
				case FINISH_ELEMENT: {
					break;
				}				
				case COMMENT: {
					packOneIndex(dest, pos, indexes, i++, type);
					break;
				}
				case NAMESPACE_DECLARATION: {
					packTwoIndexes(dest, pos, indexes, i, type);
					i += 2;
					break;
				}
				case PROCESSING_INSTRUCTION: {
					packTwoIndexes(dest, pos, indexes, i, type);
					i += 2;
					break;
				}
				case DOCTYPE: { // no need to pack indexes
					dest.putInt(indexes[i++]);
					dest.putInt(indexes[i++]);
					dest.putInt(indexes[i++]);
					dest.putInt(indexes[i++]);
					break;
				}
				case START_DOCUMENT: {
					packOneIndex(dest, pos, indexes, i++, type);
					break;
				}
				case FINISH_DOCUMENT: {
					return; // all of document has been packed
				}
				default: {
					throw new IllegalArgumentException("illegal node type");
				}
			}
		}
	}

	private static void packOneIndex(ByteBuffer dest, int pos, int[] indexes, int i, int type) {
		type |= packIndex(dest, indexes[i]) << 4;
		dest.put(pos, (byte)type);
	}
	
	private static void packTwoIndexes(ByteBuffer dest, int pos, int[] indexes, int i, int type) {
		type |= packIndex(dest, indexes[i]) << 4;
		type |= packIndex(dest, indexes[i+1]) << 6;
		dest.put(pos, (byte)type);
	}
	
	/** pack 4 byte integer index into 1, 2 or 4 bytes, depending on range */
	private static int packIndex(ByteBuffer dest, int index) {
		if (index <= 255) { 
			dest.put((byte) index); 
			return 0; // bits 00
		}
		else if (index <= 65535) { 
			dest.putShort((short) index);
			return 1; // bits 01
		}
		else {
			dest.putInt(index);
			return 3; // bits 11
		}
	}
				
	/** parse document from encoded buffer */
	private Document readDocument(ByteBuffer buf) throws BinaryParsingException {
		int type = buf.get();
		if ((type & 0x0F) != START_DOCUMENT) throw new BinaryParsingException("Missing document");
		
		// add children of document, retaining the exact same order found in input
		Document doc = new Document(new Element("fakeRoot"));
		doc.setBaseURI(readString(buf, 4, type));
		boolean hasRootElement = false;
		int i = 0;
		while (buf.hasRemaining()) {
			Node child = readNode(buf, buf.get());
			if (child == null) break; // end of document
			
			if (child instanceof Element) { // replace fake root with real root
				doc.setRootElement((Element) child); 
				hasRootElement = true;
			} else {
				doc.insertChild(child, i);
			}
			i++;
		}
		if (!hasRootElement) throw new BinaryParsingException("Missing root element");
		return doc;
	}
	
	/** parse the next node from an encoded buffer */
	private Node readNode(ByteBuffer buf, int type) throws BinaryParsingException {
		switch (type & 0x0F) {
			case TEXT: {
				return readText(buf, type);
			}
			case FREQUENT_TEXT: {
				return readFrequentText(type);
			}
			case ATTRIBUTE: {
				throw new BinaryParsingException("Unreachable attribute");
			}
			case ATTRIBUTE_WITH_NULL_NAMESPACE: {
				throw new BinaryParsingException("Unreachable attribute");
			}
			case START_ELEMENT: {
				return readElement(buf, type);
			}
			case START_ELEMENT_WITH_NULL_NAMESPACE: {
				return readElement(buf, type);
			}
			case FINISH_ELEMENT: {
				return null;
			}
			case COMMENT: {
				return readComment(buf, type);
			}
			case NAMESPACE_DECLARATION: {
				throw new BinaryParsingException("Unreachable namespace declaration");
			}
			case PROCESSING_INSTRUCTION: {
				return readProcessingInstruction(buf);
			}
			case DOCTYPE: {
				return readDocType(buf);
			}
			case START_DOCUMENT: {
				throw new BinaryParsingException("Unreachable start of document");
			}
			case FINISH_DOCUMENT: {
				return null;
			}
			default: {
				throw new BinaryParsingException("Illegal node type code=" + type);
			}
		}
	}
		
	private Element readElement(ByteBuffer buf, int type) throws BinaryParsingException {
		String namespaceURI = "";
		String qname = readString(buf, 4, type);
		if ((type & 0x0F) != START_ELEMENT_WITH_NULL_NAMESPACE) {
			namespaceURI = readString(buf, 6, type);
		}		
		Element elem = this.readElementCache.build(qname, namespaceURI);
//		Element elem = new Element(qname, namespaceURI);
		
		for (;;) {
			Node node;
			type = buf.get();
			switch (type & 0x0F) {
				case TEXT: {
					node = readText(buf, type);
					break;
				}
				case FREQUENT_TEXT: {
					node = readFrequentText(type);
					break;
				}
				case ATTRIBUTE: {
					readAttribute(buf, elem, type);
					continue;
				}
				case ATTRIBUTE_WITH_NULL_NAMESPACE: {
					readAttribute(buf, elem, type);
					continue;
				}
				case START_ELEMENT: {
					node = readElement(buf, type);
					break;
				}
				case START_ELEMENT_WITH_NULL_NAMESPACE: {
					node = readElement(buf, type);	
					break;
				}
				case FINISH_ELEMENT: {
					return elem; // done with current element; return from method
				}
				case COMMENT: {
					node = readComment(buf, type);
					break;
				}
				case NAMESPACE_DECLARATION: {
					readNamespaceDeclaration(buf, elem, type);
					continue;
				}
				case PROCESSING_INSTRUCTION: {
					node = readProcessingInstruction(buf);
					break;
				}
//				case DOCTYPE: {
//					throw new BinaryParsingException("Unreachable DocType");
//				}
//				case START_DOCUMENT: {
//					throw new BinaryParsingException("Unreachable start document");
//				}
//				case FINISH_DOCUMENT: {
//					throw new BinaryParsingException("Unreachable finish document");
//				}
				default: {
					throw new BinaryParsingException("Illegal node type");
				}
			}
			// assert node != null
			elem.appendChild(node);
		}
	}
	
	private void readNamespaceDeclaration(ByteBuffer buf, Element elem, int type) {
		String prefix = readString(buf, 4, type);
		String uri = readString(buf, 6, type);
		elem.addNamespaceDeclaration(prefix, uri);
	}
	
	private void readAttribute(ByteBuffer buf, Element elem, int type) throws BinaryParsingException {
		String namespaceURI = "";
		String qname = readString(buf, 4, type);
		String value = readString(buf, 6, type);
		Attribute.Type attrType = getAttributeType(buf.get());
		if ((type & 0x0F) != ATTRIBUTE_WITH_NULL_NAMESPACE) {
			// URI has already been read as part of namespace declaration of current element
			int i = qname.indexOf(':');
			String prefix = qname.substring(0, i);
			namespaceURI = elem.getNamespaceURI(prefix);
		}
		
		// take interned attr from cache, if available, avoiding XML reverification
		Attribute attr = this.readAttributeCache.build(qname, namespaceURI, value, attrType);
//		Attribute attr = new Attribute(qname, namespaceURI, value, attrType);
		elem.addAttribute(attr);
	}
	
	private Text readText(ByteBuffer buf, int type) {
//		return new Text(readString(buf, 4, type));
		switch ((type >>> 4) & 0x03) { // look at two bits indicating index length
			case 0 : {
				// try to avoid XML reverification by caching frequent Texts
				int i = readUnsignedByte(buf.get());
				if (textCache[i] != null) { // cache hit
//					hits++;
					return new Text(textCache[i]);
				} else { // cache miss
//					misses++;
					Text text = new Text(nameList[i]);
					textCache[i] = text;
					return text;
				}
			}
			case 1 : return new Text(nameList[readUnsignedShort(buf.getShort())]);
			default: return new Text(nameList[buf.getInt()]);
		}
	}
	
	/** use high nibble to identify code of frequent text */
	private Text readFrequentText(int type) {
		return new Text(frequentTexts[(type & 0xFF) >>> 4]);
	}

	private Comment readComment(ByteBuffer buf, int type) {
		return new Comment(readString(buf, 4, type));
	}
	
	private ProcessingInstruction readProcessingInstruction(ByteBuffer buf) {
		int type = buf.get(buf.position() - 1);
		String target = readString(buf, 4, type);
		String value = readString(buf, 6, type);
		return new ProcessingInstruction(target, value);
	}
	
	/** do not pack indexes of doctype */
	private DocType readDocType(ByteBuffer buf) {
		String rootElementName = nameList[buf.getInt()];
		String publicID = nameList[buf.getInt()];
		if (publicID.length() == 0) publicID = null;
		String systemID = nameList[buf.getInt()];
		if (systemID.length() == 0) systemID = null;
		String internalDTDSubset = nameList[buf.getInt()];
		if (internalDTDSubset.length() == 0) internalDTDSubset = null;
		
		DocType docType = new DocType(rootElementName, publicID, systemID);
		docType.setInternalDTDSubset(internalDTDSubset); // talk to Elliotte about this
		return docType;
	}
	
	/** read string via packed index from namepool */
	private String readString(ByteBuffer buf, int shift, int type) {
		// assert shift == 4 || shift == 6
		switch ((type >>> shift) & 0x03) { // look at two bits indicating index length
			case 0 : return nameList[readUnsignedByte(buf.get())];
			case 1 : return nameList[readUnsignedShort(buf.getShort())];
			default: return nameList[buf.getInt()];
		}
	}
	
	/** encode a document into the intermediate unpacked binary form */
	private void writeDocument(Document doc) {
		nodeTokens.add((byte)START_DOCUMENT);
		writeIndex(doc.getBaseURI());
		for (int i = 0; i < doc.getChildCount(); i++) {
			writeNode(doc.getChild(i));
		}
		nodeTokens.add((byte)FINISH_DOCUMENT);
	}
	
	/** encode a node into the intermediate binary form */
	private void writeNode(Node node) {
		if (node instanceof Text) {
			writeText(((Text) node));
		} else if (node instanceof Element) {
			writeElement(((Element) node));
		} else if (node instanceof Comment) {
			writeComment(((Comment) node));
		} else if (node instanceof ProcessingInstruction) {
			writeProcessingInstruction(((ProcessingInstruction) node));
		} else if (node instanceof DocType) {
			writeDocType(((DocType) node));
		} else {
			throw new IllegalArgumentException("Unknown node type (should never happen)");
		}
	}

	private void writeAttribute(Attribute attr) {
		int type = ATTRIBUTE_WITH_NULL_NAMESPACE;
		String qname = attr.getLocalName();
		if (attr.getNamespacePrefix().length() != 0) {
			type = ATTRIBUTE;
			qname = this.writeAttributeCache.intern(qname, attr.getNamespacePrefix());
//			qname = prefix + ":" + qname;
		}
		
		writeIndex(qname);
//		writeIndex(attr.getQualifiedName());
		writeIndex(attr.getValue());
		nodeTokens.add((byte)type);
		nodeTokens.add(getAttributeTypeCode(attr));
	}
	
	private void writeElement(Element elem) {
		writeIndex(elem.getQualifiedName());
		int type = START_ELEMENT_WITH_NULL_NAMESPACE;
		if (elem.getNamespaceURI().length() != 0) {
			type = START_ELEMENT;
			writeIndex(elem.getNamespaceURI());
		}
		nodeTokens.add((byte)type);
		
		writeNamespaceDeclarations(elem);
		
		int count = elem.getAttributeCount();
		for (int i = 0; i < count; i++) {
			writeAttribute(elem.getAttribute(i));
		}
		
		count = elem.getChildCount();
		for (int i = 0; i < count; i++) {
			writeNode(elem.getChild(i));
		}
		
		nodeTokens.add((byte)FINISH_ELEMENT);
	}

	private void writeNamespaceDeclarations(Element elem) {
		int count = elem.getNamespaceDeclarationCount();
		if (count == 1) 
			return; // elem.getNamespaceURI() has already been written

		for (int i = 0; i < count; i++) {
			String prefix = elem.getNamespacePrefix(i);
			String uri = elem.getNamespaceURI(prefix);
			if (prefix.equals(elem.getNamespacePrefix()) && uri.equals(elem.getNamespaceURI())) {
//				if (DEBUG) System.err.println("********** NAMESPACE IGNORED ON WRITE ***************\n");
				continue;
			}
			nodeTokens.add((byte)NAMESPACE_DECLARATION);
			writeIndex(prefix);
			writeIndex(uri);
		}
	}
	
	private void writeText(Text text) {
		String value = text.getValue();
		int code = getFrequentTextCode(value);
		if (code >= 0) {
			nodeTokens.add((byte) (FREQUENT_TEXT | (code << 4)));
		} else {
			nodeTokens.add((byte)TEXT);
			writeIndex(value);
		}
	}
	
	private void writeComment(Comment comment) {
		nodeTokens.add((byte)COMMENT);
		writeIndex(comment.getValue());
	}
	
	private void writeProcessingInstruction(ProcessingInstruction pi) {
		nodeTokens.add((byte)PROCESSING_INSTRUCTION);
		writeIndex(pi.getTarget());
		writeIndex(pi.getValue());
	}
	
	/** do not pack indexes of doctype, unnecessary anyway, since infrequent */
	private void writeDocType(DocType docType) {
		nodeTokens.add((byte)DOCTYPE);
		writeIndex(docType.getRootElementName());
		writeIndex(docType.getPublicID() == null ? "" : docType.getPublicID());
		writeIndex(docType.getSystemID() == null ? "" : docType.getSystemID());
		writeIndex(docType.getInternalDTDSubset() == null ? "" : docType.getInternalDTDSubset());
	}	
	
	/** put name into namepool and append the string's index to indexData */
	private final void writeIndex(String name) {
		StringInfo info = (StringInfo) namePool.get(name);
		int index;
		if (info == null) { // name hasn't been seen before; add it
			index = name.indexOf(NIL); // sanity check
			if (index >= 0) throwNILCharException(name, index);
			index = namePool.size();
			namePool.put(name, new StringInfo(index));
		} else { // already existing name occurs yet again
			info.frequency++;
			index = info.index;
		}
		indexData.add(index);
	}
	
	private static void throwNILCharException(String name, int index) {
		throw new IllegalArgumentException(
			"Name must not contain NIL char 0x00 [index:" + index + " within '" 
			+ name + "']");
	}

	/** reuse XOM's way of representing attribute types */
	private static byte getAttributeTypeCode(Attribute attr) {
		return (byte) attr.getType().hashCode();
	}

	private static Attribute.Type getAttributeType(int typeCode) throws BinaryParsingException {
		switch (typeCode) {
			case 0 : return Attribute.Type.UNDECLARED;
			case 1 : return Attribute.Type.CDATA;
			case 2 : return Attribute.Type.ID;
			case 3 : return Attribute.Type.IDREF;
			case 4 : return Attribute.Type.IDREFS;
			case 5 : return Attribute.Type.NMTOKEN;
			case 6 : return Attribute.Type.NMTOKENS;
			case 7 : return Attribute.Type.NOTATION;
			case 8 : return Attribute.Type.ENTITY;
			case 9 : return Attribute.Type.ENTITIES;
			case 10 : return Attribute.Type.ENUMERATION;
			default: 
				throw new BinaryParsingException("Illegal attribute type code: " + typeCode);
		}
	}

	// short texts that will be compressed into particularly compact form; 
	// max 16 values (4 bits); mostly line/element separator-type strings
	private static final Text[] frequentTexts = { 
		/* 0 */ new Text(""),
		/* 1 */ new Text("\n"), new Text(" "), // "\r", 
		/* 2 */ new Text("\r\n"), new Text("\n\n"), // "  ",
		/* 3 */ new Text("\n  "), // "   ", "\r\n ",
		/* 4 */ new Text("\r\n  "), new Text("\r\n\r\n"), // "    ", , "\n   ",
		/* 5 */ new Text("\n    "),
		/* 6 */ new Text("\r\n    ")
	};

	/** is val a frequent string? if so which one? */
	private static int getFrequentTextCode(String val) {
		switch (val.length()) {
			case 0 : {
				return 0;
			}
			case 1 : {
				char c = val.charAt(0);
				if (c == '\n') return 1;
				else if (c == ' ') return 2;
				break;
			}
			case 2 : {
				char c1 = val.charAt(0);
				char c2 = val.charAt(1);
				if (c1 == '\r' && c2 == '\n') return 3;
				else if (c1 == '\n' && c2 == '\n') return 4;
				break;
			}
			case 3 : {
				if ("\n  ".equals(val)) return 5;
				break;
			}
			case 4 : {
				if ("\r\n  ".equals(val)) return 6;
				else if ("\r\n\r\n".equals(val)) return 7;
				break;
			}
			case 5 : {
				if ("\n    ".equals(val)) return 8;
				break;
			}
			case 6 : {
				if ("\r\n    ".equals(val)) return 9;
				break;
			}
		}
		return -1; // not a frequent text
	}

	/** ZLIB compression */
	private byte[] doDeflate(ArrayByteList src, int compressionLevel) {
		if (deflater == null || compressionLevel != deflationLevel) {
			if (deflater != null) deflater.end(); // free resources
			deflater = new Deflater(compressionLevel);
			deflationLevel = compressionLevel;
		}
		deflater.reset();
		deflater.setInput(src.asArray(), 0, src.size());
		int minBufSize = 256;
		byte[] result = new byte[Math.max(minBufSize, src.size() / 3)];
		int size = 0;
		do {
			if (result.length - size < minBufSize) {
				byte[] tmp = new byte[2 * result.length];
				System.arraycopy(result, 0, tmp, 0, size);
				result = tmp;
			}
			size += deflater.deflate(result, size, result.length - size);
		} while (!deflater.needsInput());
		
		if (!deflater.finished()) {
		    deflater.finish();
		    while (!deflater.finished()) {
				if (result.length - size < minBufSize) {
					byte[] tmp = new byte[2 * result.length];
					System.arraycopy(result, 0, tmp, 0, size);
					result = tmp;
				}
				size += deflater.deflate(result, size, result.length - size);
		    }
		}
		
		deflater.reset();
		byte[] out = new byte[size];
		System.arraycopy(result, 0, out, 0, size);
		return out;
	}
	
	/** ZLIB decompression */
	private ByteBuffer doInflate(byte[] bnuxDocument) throws BinaryParsingException {
		if (inflater == null) inflater = new Inflater();
		inflater.reset();
		inflater.setInput(bnuxDocument);

		int minBufSize = 256;
		byte[] result = new byte[Math.max(minBufSize, 3*bnuxDocument.length)];
		int size = 0;
		do {
			if (result.length - size < minBufSize) {
				byte[] tmp = new byte[2 * result.length];
				System.arraycopy(result, 0, tmp, 0, size);
				result = tmp;
			}
			try {
				size += inflater.inflate(result, size, result.length - size);
			} catch (DataFormatException e) {
				String s = e.getMessage();
			    throw new BinaryParsingException(s != null ? s : "Invalid ZLIB data format", e);
			}
		} while (!(inflater.finished() || inflater.needsDictionary()));
		
		inflater.reset();
		return ByteBuffer.wrap(result, 0, size);	
	}

	/**
	 * Returns the integer value gained when interpreting the value v as an <em>unsigned byte</em>.
	 * Example: readUnsignedByte(0xFF) == 255, rather than -1.
	 */
	private static int readUnsignedByte(byte v) {
		return v & 0xFF;
	}

	/**
	 * Returns the integer value gained when interpreting the value v as an <em>unsigned short</em>.
	 */
	private static int readUnsignedShort(short v) {
		return v & 0xFFFF;
	}


	///////////////////////////////////////////////////////////////////////////////
	// Nested classes:
	///////////////////////////////////////////////////////////////////////////////
	
	/** A value in the namepool HashMap */
	private static final class StringInfo {
		public int index;     // index to correlate nameinfo with indexList
		public int frequency; // number of occurances of String within current XML document
		
		public StringInfo(int index) { 
			this.index = index; 
			this.frequency = 1; 
		}
		
		public String toString() { // DEBUG and PROFILE ONLY
			return "freq@" + frequency;
		}
	}
	
	/** 
	 * Small fast cache for the last N attributes. 
	 * Used by deserialization. 
	 * Avoids repeated XML reverification in XOM Attribute constructor on cache hit,
	 * plus avois generating new String objects by "interning" localName and prefix.
	 * Has overhead close to zero on cache miss. 
	 */
	private static final class ReadAttributeMicroCache {
		
		private static final int SIZE = 32;
		private final String[] qnames = new String[SIZE];
		private final String[] uris = new String[SIZE];
		private final Attribute[] attributes = new Attribute[SIZE];
		private int cursor = SIZE - 1;

		public void reset() {
			cursor = SIZE - 1;
		}
		
		public Attribute build(String qname, String uri, String value, Attribute.Type attrType) {
			for (int i = qnames.length; --i >= 0; ) {
				if (qname == qnames[i] && uri == uris[i]) { // cache hit
					Attribute attr = new Attribute(attributes[i]);
					if (attr.getType() != attrType) attr.setType(attrType);
					attr.setValue(value);
//					hits++;
					return attr;
				}
			}
			
			// cache miss
			return put(qname, uri, value, attrType);
		}

		private Attribute put(String qname, String uri, String value, Attribute.Type attrType) {
//			misses++;
			Attribute attr = new Attribute(qname, uri, value, attrType);
			qnames[cursor] = qname;
			uris[cursor] = uri;
			attributes[cursor] = attr;		
			if (cursor == 0) cursor = SIZE; // wrap around
			cursor--;
			return new Attribute(attr);
		}
	}
	
	/** 
	 * Small fast cache for the last N attributes, caching interned qnames. 
	 * Used by serialization. 
	 */
	private static final class WriteAttributeMicroCache {
		
		private static final int SIZE = 32;
		private final String[] localNames = new String[SIZE];
		private final String[] prefixes = new String[SIZE];
		private final String[] qnames = new String[SIZE];
		private int cursor = SIZE - 1;

		public void reset() {
			cursor = SIZE - 1;
		}
		
		public String intern(String localName, String prefix) {
			for (int i = localNames.length; --i >= 0; ) {
				if (localName == localNames[i] && prefix == prefixes[i]) { // cache hit
					return qnames[i];
				}
			}
			// cache miss
			return put(localName, prefix);
		}

		private String put(String localName, String prefix) {
			String qname = prefix + ":" + localName;
			localNames[cursor] = localName;
			prefixes[cursor] = prefix;
			qnames[cursor] = qname;		
			if (cursor == 0) cursor = SIZE; // wrap around
			cursor--;
			return qname;
		}
	}
	
	/** 
	 * Small fast cache for the last N elements.
	 */  
	private static final class ReadElementMicroCache {
		
		private static final int SIZE = 32;
		private final String[] qnames = new String[SIZE];
		private final String[] uris = new String[SIZE];
		private final Element[] elems = new Element[SIZE];
		private int cursor;
		private int hits;   // stats
		private int misses; // stats
		
		public ReadElementMicroCache() {
			reset();
		}

		public void reset() {
			hits = 0;
			misses = 0;
			cursor = SIZE - 1;
		}
		
		public Element build(String qname, String uri) {
			// if the cache is not working well, let's stop using it, avoiding its overhead
			if (misses > hits && hits + misses > 74) { // pragmatic heuristic
				return new Element(qname, uri);
			}
			return get(qname, uri);
		}
			
		private Element get(String qname, String uri) {
			hits++;
			for (int i = qnames.length; --i >= 0; ) {
				if (qname == qnames[i] && uri == uris[i]) { // cache hit
					return new Element(elems[i]);
				}
			}
			
			// cache miss
			return put(qname, uri);
		}
		
		private Element put(String qname, String uri) {
			hits--;
			misses++;
			Element elem = new Element(qname, uri);
			qnames[cursor] = qname;
			uris[cursor] = uri;
			elems[cursor] = elem;
			if (cursor == 0) cursor = SIZE; // wrap around
			cursor--;
			return new Element(elem);
		}
				
		public String toString() { // DEBUG and PROFILE ONLY
			return "hits=" + hits + ", misses="+misses;
		}
	}
	
}