/*
 * Copyright (c) 2003, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.pool;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;

import javax.xml.bind.JAXBException;
import javax.xml.bind.MarshalException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.UnmarshalException;
import javax.xml.bind.Unmarshaller;
import javax.xml.parsers.DocumentBuilder;

import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.NodeFactory;
import nu.xom.Nodes;
import nu.xom.ParentNode;
import nu.xom.Serializer;
import nu.xom.XMLException;
import nu.xom.canonical.Canonicalizer;
import nu.xom.converters.DOMConverter;

import org.w3c.dom.DOMImplementation;

/**
 * Various utilities avoiding redundant code in several classes.
 * 
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.67 $, $Date: 2005/02/10 00:49:35 $
 */
public class XOMUtil {
	
	// trick to detect default platform charset
	private static final Charset DEFAULT_PLATFORM_CHARSET = 
		Charset.forName(new InputStreamReader(new ByteArrayInputStream(new byte[0])).getEncoding());	

	// constructing a new DocumentBuilder is VERY expensive, so we reuse it time and again
	// in a thread-safe way once it has been constructed.
	private static final ThreadLocal threadLocalDocBuilder = new ThreadLocal() {
		protected Object initialValue() { // lazy init
			javax.xml.parsers.DocumentBuilderFactory factory =
				javax.xml.parsers.DocumentBuilderFactory.newInstance();
			factory.setNamespaceAware(true);
			try {
				DocumentBuilder docBuilder = factory.newDocumentBuilder();
				//System.err.println("found DOM DocumentBuilder="+docBuilder.getClass().getName());
				return docBuilder;
			} catch (javax.xml.parsers.ParserConfigurationException e) {
				throw new XMLException("Can't find or create DOM DocumentBuilder", e);
			}
		}
	};
	
	// in toPrettyXML(), reusing the serializer in a thread-safe way is 5 times 
	// faster for tiny documents (due too setup cost of streams, encoders, etc.)
	private static final ThreadLocal threadlocalSerializer = new ThreadLocal() {
		protected Object initialValue() { // lazy init
			ResetableByteArrayOutputStream out = new ResetableByteArrayOutputStream(128);
			try {
				Serializer serializer = new Serializer(out, "UTF-8");
				serializer.setIndent(4);
				return new Object[] { serializer, out };
			} catch (IOException e) { // safe: UTF-8 support is required by JDK spec
				throw new RuntimeException("should never happen", e);
			}			
		}
	};
	
	/** 
	 * Enables giving excessive memory back to the garbage collector, 
	 * preventing "hidden" memory leaks. 
	 */
	private static final class ResetableByteArrayOutputStream extends ByteArrayOutputStream {
		public ResetableByteArrayOutputStream(int initialCapacity) { 
			super(initialCapacity); 
		}
		public synchronized void reset(int maxCapacity) {
			reset();
			if (buf.length > maxCapacity) buf = new byte[maxCapacity]; 
		}
	}
		
	private XOMUtil() {} // not instantiable

	/**
	 * Returns a DOMImplementation via the default JAXP lookup mechanism.
	 */
	public static DOMImplementation getDOMImplementation() {
		return ((DocumentBuilder) threadLocalDocBuilder.get()).getDOMImplementation();
	}
	
	/**
	 * Reads until end-of-stream and returns all read bytes, finally closes the stream.
	 * 
	 * @param input the input stream
	 * @throws IOException if an I/O error occurs while reading the stream
	 */
	public static byte[] toByteArray(InputStream input) throws IOException {
		try {
			// safe and fast even if input.available() behaves weird or buggy
			int len = Math.max(256, input.available());
			byte[] buffer = new byte[len];
			byte[] output = new byte[len];
			
			len = 0;
			int n;
			while ((n = input.read(buffer)) >= 0) {
				if (len + n > output.length) { // grow capacity
					byte tmp[] = new byte[Math.max(output.length << 1, len + n)];
					System.arraycopy(output, 0, tmp, 0, len);
					System.arraycopy(buffer, 0, tmp, len, n);
					buffer = output; // use larger buffer for future larger bulk reads
					output = tmp;
				} else {
					System.arraycopy(buffer, 0, output, len, n);
				}
				len += n;
			}

			if (len == output.length) return output;
			buffer = null; // help gc
			buffer = new byte[len];
			System.arraycopy(output, 0, buffer, 0, len);
			return buffer;
		} finally {
			if (input != null) input.close();
		}
	}

	/**
	 * Reads until end-of-stream and returns all read bytes as a string, finally
	 * closes the stream, converting the data with the given charset encoding, or
	 * the system's default platform encoding if <code>charset == null</code>.
	 * 
	 * @param input the input stream
	 * @param charset the charset to convert with, e.g. <code>Charset.forName("UTF-8")</code>
	 * @throws IOException if an I/O error occurs while reading the stream
	 */
	public static String toString(InputStream input, Charset charset) throws IOException {
		if (charset == null) charset = DEFAULT_PLATFORM_CHARSET;			
		byte[] data = toByteArray(input);
		return charset.decode(ByteBuffer.wrap(data)).toString();
	}

	/**
	 * Returns a pretty-printed String representation of the given node (subtree).
	 * 
	 * @param node the node (subtree) to convert.
	 * @see Serializer
	 */
	public static String toPrettyXML(Node node) {
		Document doc;
		if (node instanceof Document) {
			doc = (Document) node;
		}
		else if (node instanceof Element) {
			// do not modify elem's parent pointer
			doc = new Document((Element) node.copy()); 
		}
		else {
			// not really pretty-printed
			return node.toXML(); 
		}

		Object[] pair = (Object[]) threadlocalSerializer.get();
		Serializer serializer = (Serializer) pair[0];
		ResetableByteArrayOutputStream out = (ResetableByteArrayOutputStream) pair[1];
		String str;
		try {
			serializer.write(doc);
			serializer.flush();
			str = out.toString("UTF-8"); // safe: UTF-8 support is required by JDK spec
		} catch (IOException e) {
			throw new RuntimeException("should never happen", e);
		} finally {
			out.reset(1024); // avoid hidden memory leaks
		}
		
		// remove XML declaration header <?xml version="1.0" encoding="UTF-8"?>\r\n
		// remove trailing line break, if any
		int i = str.indexOf('\n');
		int j = str.length();
		if (str.endsWith("\r\n")) j = j - 2;
		else if (str.endsWith("\n")) j = j - 1;
		return str.substring(i+1, j);
	}
	
	/**
	 * Returns the W3C Canonical XML representation of the given document.
	 * 
	 * @param doc the document to convert.
	 * @see Canonicalizer
	 */
	public static byte[] toCanonicalXML(Document doc) {
		ByteArrayOutputStream out = new ByteArrayOutputStream(128);
		try {
			new Canonicalizer(out).write(doc);
		} catch (IOException e) {
			throw new RuntimeException("should never happen", e);
		}
		return out.toByteArray();
	}
	
	/**
	 * Returns a node factory that removes each {@link nu.xom.Text} node that
	 * is empty or consists of whitespace characters only. This method fully preserves
	 * narrative <code>Text</code> containing whitespace along with other characters.
	 * <p>
	 * Otherwise this factory behaves just like the standard {@link NodeFactory}.
	 * <p>
	 * Ignoring whitespace-only nodes reduces memory footprint for documents
	 * that are heavily pretty printed and indented, i.e. human-readable. Remember that
	 * without such a factory, <i>every </i> whitespace sequence occurring
	 * between element tags generates a mostly useless Text node.
	 * <p>
	 * Finally, note that this method's whitespace pruning is appropriate for
	 * many, but not all XML use cases (round-tripping).
	 * For example, the blank between 
	 * <code>&lt;p>&lt;strong>Hello&lt;/strong> &lt;em>World!&lt;/em>&lt;/p></code>
	 * will be removed, which might not be what you want. This is because this 
	 * method does not look across multiple Text nodes.
	 */
	public static NodeFactory getIgnoreWhitespaceOnlyTextNodeFactory() {
		return new NodeFactory() {
			private final Nodes NONE = new Nodes();
			
			public Nodes makeText(String text) {
				for (int i=text.length(); --i >= 0; ) {
					if (text.charAt(i) > ' ') // see String.trim() implementation 
						return super.makeText(text); 					
//					if (! Character.isWhitespace(text.charAt(i))) 
//						return super.makeText(text);
				}
				return NONE;
			}			
		};
	}
	
	/**
	 * Returns a node factory that removes leading and trailing whitespaces in
	 * each {@link nu.xom.Text} node, altogether removing a Text node that becomes empty after
	 * said trimming. Trimming is done via {@link String#trim()}. For example a
	 * text node of <code>"  hello world   "</code> becomes
	 * <code>"hello world"</code>, and a text node of <code>"   "</code> is
	 * removed.
	 * <p>
	 * Otherwise this factory behaves just like the standard {@link NodeFactory}.
	 * <p>
	 * Finally, note that this method's whitespace pruning is appropriate for
	 * many, but not all XML use cases (round-tripping).
	 */
	public static NodeFactory getTextTrimmingNodeFactory() {
		return new NodeFactory() {
			private final Nodes NONE = new Nodes();
			
			public Nodes makeText(String text) {
				text = text.trim();
				return text.length() == 0 ? NONE : super.makeText(text);
			}
		};
	}
	
	/**
	 * Returns a node factory for pure document validation. This factory does
	 * not generate a document on <code>Builder.build(...)</code>, which is
	 * not required anyway for pure validation. Ignores all input and builds an
	 * empty document instead. This improves validation performance.
	 */
	public static NodeFactory getNullNodeFactory() {
		return new NodeFactory() {		
			private final Nodes NONE = new Nodes();
			
			public Nodes makeAttribute(String name, String URI, String value, Attribute.Type type) {
				return NONE;
			}
	
			public Nodes makeComment(String data) {
				return NONE;
			}
	
			public Nodes makeDocType(String rootElementName, String publicID, String systemID) {
				return NONE;
			}
	
			public Nodes makeProcessingInstruction(String target, String data) {
				return NONE;
			}
			
			public Element makeRootElement(String name, String namespace) {
				return new Element(name, namespace);
			}
	
			public Nodes makeText(String text) {
				return NONE;
			}
	
			public Element startMakingElement(String name, String namespace) {
				return null;
			}
		};
	}

	/**
	 * Marshals (serializes) the given JAXB object via the given marshaller
	 * into a new XOM Document (convenience method).
	 * <p>
	 * This implementation is somewhat inefficient but correctly does the job.
	 * There is no connection between the JAXB object tree and the XOM object tree;
	 * they are completely independent object trees without any cross-references.
	 * Hence, updates in one tree are not automatically reflected in the other tree.
	 * 
	 * @param marshaller
	 *            a JAXB serializer (note that a marshaller is typically
	 *            not thread-safe and expensive to construct; hence the recommendation 
	 *            is to use a {@link ThreadLocal} to make it thread-safe and efficient)
	 * @param jaxbObj
	 *            the JAXB object to serialize
	 * @return the new XOM document
	 * @throws JAXBException
	 *             If an unexpected problem occurred in the conversion.
	 * @throws MarshalException
	 *             If an error occurred while performing the marshal operation.
	 *             Whereever possible, one should prefer the
	 *             {@link MarshalException} over the {@link JAXBException}.
	 * @see Marshaller#marshal(java.lang.Object, org.w3c.dom.Node)
	 */
	public static Document jaxbMarshal(Marshaller marshaller, Object jaxbObj) 
			throws JAXBException {
		if (jaxbObj == null) 
			throw new IllegalArgumentException("jaxbObj must not be null");
		if (marshaller == null) 
			throw new IllegalArgumentException("marshaller must not be null");
		
		return DOMConverter.convert(jaxbMarshalDOM(marshaller, jaxbObj));
	}
	
	/**
	 * Unmarshals (deserializes) the given XOM node via the given unmarshaller
	 * into a new JAXB object (convenience method).
	 * <p>
	 * This implementation is somewhat inefficient but correctly does the job.
	 * There is no connection between the JAXB object tree and the XOM object tree;
	 * they are completely independent object trees without any cross-references.
	 * Hence, updates in one tree are not automatically reflected in the other tree.
	 * 
	 * @param unmarshaller
	 *            a JAXB deserializer (note that an unmarshaller is typically
	 *            not thread-safe and expensive to construct; hence the recommendation 
	 *            is to use a {@link ThreadLocal} to make it thread-safe and efficient)
	 * @param node
	 *            the XOM node to deserialize
	 * @return the new JAXB object
	 * @throws JAXBException
	 *             If an unexpected problem occurred in the conversion.
	 * @throws UnmarshalException
	 *             If an error occurred while performing the unmarshal operation.
	 *             Whereever possible, one should prefer the
	 *             {@link UnmarshalException} over the {@link JAXBException}.
	 * @see Unmarshaller#unmarshal(org.w3c.dom.Node)
	 */
	public static Object jaxbUnmarshal(Unmarshaller unmarshaller, ParentNode node)
			throws JAXBException {
		if (node == null) 
			throw new IllegalArgumentException("node must not be null");
		if (unmarshaller == null) 
			throw new IllegalArgumentException("unmarshaller must not be null");
		
		Document doc;
		if (node instanceof Document) {
			doc = (Document) node;
		}
		else if (node instanceof Element) {
			// do not modify elem's parent pointer
			doc = new Document((Element) node.copy());
		}
		else {
			throw new IllegalArgumentException("Illegal XOM node type" + node);
		}
		
		return jaxbUnmarshalDOM(unmarshaller, 
				DOMConverter.convert(doc, getDOMImplementation()));
	}
	
	private static org.w3c.dom.Document jaxbMarshalDOM(Marshaller marshaller, Object jaxbObj) throws JAXBException {
		DocumentBuilder docBuilder = (DocumentBuilder) threadLocalDocBuilder.get();
		org.w3c.dom.Document doc = docBuilder.newDocument();
		marshaller.marshal(jaxbObj, doc);
		return doc;
	}
	
	private static Object jaxbUnmarshalDOM(Unmarshaller unmarshaller, org.w3c.dom.Node node) throws JAXBException {
		return unmarshaller.unmarshal(node);
	}
	
//	/**
//	 * Returns the XOM document obtained by parsing the given XML string.
//	 * Useful for quick'n dirty inline examples and tests.
//	 * The document is parsed with a non-validating Builder, 
//	 * and the baseURI of the document will be the empty string.
//	 * <p>
//	 * Example usage:
//	 * <pre>
//	 * String xml = 
//	 *     "&lt;foo>" +
//	 *         "&lt;bar size=\"123\">" +
//	 *         "&lt;/bar>" +
//	 *     "&lt;/foo>";
//	 * Document doc = toDocument(xml);
//	 * </pre>
//	 * 
//	 * @param xml the string to parse from
//	 * @return the corresponding XOM document
//	 * @throws XMLException if the string is not well-formed XML.
//	 * @see nu.xom.Builder#build(String, String)
//	 */
//	public static Document toDocument(String xml) {
//		try { // no need to be inefficient by default
//			return BuilderPool.GLOBAL_POOL.getBuilder(false).build(xml, "");
//		} catch (ValidityException e) {
//			throw new XMLException(e.getMessage(), e);
//		} catch (ParsingException e) {
//			throw new XMLException(e.getMessage(), e);
//		} catch (IOException e) {
//			throw new XMLException(e.getMessage(), e);
//		}
//	}
			
}