/*
 * Copyright (c) 2003, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.tests;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;

import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Serializer;
import nux.xom.binary.BinaryXMLCodec;
import nux.xom.pool.XOMUtil;

/** 
 * Command-line demo that converts a set of standard textual XML files to and from binary xml (bnux) files;
 * Bnux files are indicated by a ".bnux" file name extension.
 * Output files are written in the same directory as the input files, 
 * with the ".bnux" file extension added or stripped, respectively. 
 * <p>
 * Example usage:
 * 
 * <pre>
 * export CLASSPATH=lib/nux.jar:lib/saxon8.jar:lib/saxon8-xom.jar:lib/xom.jar
 * 
 * java -server nux.xom.tests.BinaryXMLConverter samples/*.xml
 * java -server nux.xom.tests.BinaryXMLConverter samples/*.bnux
 * </pre>
 * 
 * @see nux.xom.binary.BinaryXMLCodec
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.2 $, $Date: 2005/02/12 21:27:49 $
 */
public final class BinaryXMLConverter {
	
	private BinaryXMLConverter() {}

	/**
	 * Runs the demo program. 
	 */
	public static void main(String[] args) throws Exception {
		// some non-standard options for some more in-depth testing:
		int compressionLevel = Integer.getInteger("nux.xom.tests.BinaryXMLConverter.compressionLevel", 0).intValue();
		int runs = Integer.getInteger("nux.xom.tests.BinaryXMLConverter.runs", 1).intValue();
		boolean readOnly = Boolean.getBoolean("nux.xom.tests.BinaryXMLConverter.readOnly");
		
		BinaryXMLCodec codec = new BinaryXMLCodec();
		Builder builder = new Builder();
		
		for (int run=0; run < runs; run++) {
			long s = System.currentTimeMillis();
			for (int i=0; i < args.length; i++) {
				long start, end;
				String fileName = args[i];
				if (new File(fileName).isDirectory()) continue; // ignore
				System.out.print("Reading " + fileName);
				
				if (fileName.endsWith(".bnux")) {
					// read
					start = System.currentTimeMillis();
					byte data[] = XOMUtil.toByteArray(new FileInputStream(fileName));
					Document doc = codec.deserialize(data);
					end = System.currentTimeMillis();
					System.out.print(" [ms=" + (end-start) + "]. ");
					if (readOnly) {
						System.out.println();
						continue;
					}
					
					// write
					String destFileName = fileName.substring(0, fileName.length() - ".bnux".length());
					System.out.print("Writing " + destFileName);
					start = System.currentTimeMillis();
					
					FileOutputStream out = new FileOutputStream(destFileName);
					Serializer ser = new Serializer(out);
					ser.write(doc);
					out.close();
					
					end = System.currentTimeMillis();
					System.out.println(" [ms=" + (end-start) + "]. ");
				}
				else {
					// read
					start = System.currentTimeMillis();
					Document doc = builder.build(new File(fileName));
					end = System.currentTimeMillis();
					System.out.print(" [ms=" + (end-start) + "]. ");
					if (readOnly) {
						System.out.println();
						continue;
					}
					
					// write
					String destFileName = fileName + ".bnux";
					System.out.print("Writing " + destFileName);
					start = System.currentTimeMillis();
					
					byte data[] = codec.serialize(doc, compressionLevel);
					FileOutputStream out = new FileOutputStream(destFileName);
					out.write(data);
					out.flush();
					out.close();
					
					end = System.currentTimeMillis();
					System.out.println(" [ms=" + (end-start) + "]. ");
				}
			}
			long e = System.currentTimeMillis();
			System.out.println("Completed run=" + run + " [ms=" + (e-s) + "].");
			System.out.println("\n");
		}
	}

}