/*
 * Copyright (c) 2003, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.tests;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Arrays;

import javax.xml.transform.stream.StreamSource;

import net.sf.saxon.Configuration;
import net.sf.saxon.query.StaticQueryContext;
import nu.xom.Builder;
import nu.xom.DocType;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Serializer;
import nux.xom.binary.BinaryXMLCodec;
import nux.xom.pool.BuilderPool;
import nux.xom.pool.XOMUtil;

/**
 * Tests and benchmarks bnux format against a set of standard textual XML files.
 * Command line options allow to round-trip and check canonical XML results.
 * They also allow to benchmark serialization, deserialization or both, 
 * either via bnux or xom.
 * <p>
 * Example usage: 
 * <pre>
 * java -server nux.xom.tests.BinaryXMLTest check bnux 0 1 data/*.xml
 * java -server nux.xom.tests.BinaryXMLTest ser bnux 0 5000 data/soap.xml
 * java -server nux.xom.tests.BinaryXMLTest deser bnux 0 5000 data/soap.xml
 * java -server nux.xom.tests.BinaryXMLTest serdeser bnux 0 5000 data/soap.xml
 * java -server nux.xom.tests.BinaryXMLTest serdeser xom 0 5000 data/soap.xml
 * </pre>
 * 
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.17 $, $Date: 2005/02/10 00:51:58 $
 */
public final class BinaryXMLTest {
	
	private BinaryXMLTest() {}
	
	public static void main(final String args[]) throws Exception {
		
		final String cmd = args[0]; // ser|deser|serdeser|check
		final String mode = args[1]; // xom|bnux|saxon
		int compressionLevel = Integer.parseInt(args[2]); // 0..9
		final int runs = Integer.parseInt(args[3]); // 1..infinity
		final boolean enablePerformancePatches = "true".equals(
			System.getProperty("nux.xom.tests.BinaryXMLTest.enablePatches", "true"));
		System.out.println("patchesEnabled=" + enablePerformancePatches);
		final boolean testCompressionLevels = false;

		for (int k=0; k < 1; k++) { // make k large to stress test for a couple of hours 
			long time = 0;
			long doneEncoded = 0;
			long done = 0;
			int checksum = 0;
			
			if (enablePerformancePatches) { // init before BinaryXMLCodec
				// temporary (?) performance hack via patch: disable some expensive sanity checks 
				System.setProperty("nu.xom.Verifier.checkPCDATA", "false");
				System.setProperty("nu.xom.Verifier.checkURI", "false");
			}
			
			BinaryXMLCodec codec = new BinaryXMLCodec();
			Builder builder = new Builder(false);
//			Builder builder = BuilderPool.GLOBAL_POOL.getBuilder(false);
			StaticQueryContext context = null;
			if (mode.equals("saxon")) context = new StaticQueryContext(new Configuration());
			
			for (int j=4; j < args.length; j++) {
				if (ignore(args[j])) continue;
				File file = new File(args[j]);
				if (file.isDirectory()) continue;
				System.out.println("now processing " + file);
//				System.out.print(".");
				
				Document doc = builder.build(file);
				byte[] data = codec.serialize(doc, compressionLevel);
				if (! cmd.equals("check")) doc = codec.deserialize(data); // use "interned" strings
				byte[] fileData = XOMUtil.toByteArray(new FileInputStream(file));
				long fileLength = file.length();
				int encodedSize = 0;
				
				long start = System.currentTimeMillis();
				for (int i = 0; i < runs; i++) {
					try {
						// serialize
						if (cmd.equals("ser") || cmd.equals("serdeser") || cmd.equals("check")) {
							if (mode.equals("xom")) {
								data = serializeWithXOM(doc);
							} else {
								data = codec.serialize(doc, compressionLevel);
							}
							checksum += data.length;
						}
						encodedSize = data.length;
						doneEncoded += encodedSize;
						
						// deserialize
						Document doc2 = null;
						if (cmd.equals("deser") || cmd.equals("serdeser") || cmd.equals("check")) {
							if (mode.equals("xom")) {
								doc2 = builder.build(new ByteArrayInputStream(fileData));
							}
							else if (mode.equals("saxon")) { // just for deser comparison
								context.buildDocument(new StreamSource(new ByteArrayInputStream(fileData)));
								doc2 = new Document(new Element("dummy"));
							} else {
								doc2 = codec.deserialize(data);
							}
							checksum += doc2.getBaseURI().length();				
						}
						
						// check correctness
						if (cmd.equals("check")) {
							if (! Arrays.equals(XOMUtil.toCanonicalXML(doc), XOMUtil.toCanonicalXML(doc2))) {
								System.err.println("Canonical XML Mismatch: ");
								System.err.println("in: " + doc.toXML());
								System.err.println("out: " + doc2.toXML());
								System.exit(0);
							}
							if (!equalsDocTypeEquals(doc.getDocType(), doc2.getDocType())) {	
								System.err.println("DocType Mismatch: ");
								System.err.println("in: " + doc.toXML());
								System.err.println("out: " + doc2.toXML());
								System.exit(0);
							}
						}
					} catch (RuntimeException e) {
						System.err.println("FATAL ERROR: " + e);
						System.err.println("in: " + doc.toXML());
						throw e;
					}
					done += fileLength;
				}
				time += (System.currentTimeMillis() - start);
				float cfactor = fileLength * 1.0f / encodedSize;
				System.out.println("compression factor = " + cfactor);			
				if (testCompressionLevels) // alternate on each iteration
					compressionLevel = (compressionLevel + 1) % 10;
			}
					
			
			System.out.println("\n****** SUMMARY ******");
			System.out.println("files = " + runs * (args.length - 4));
			System.out.println("secs = " + (time / 1000.0f));
			System.out.println("mean throughput MB/s = "
					+ ((done / (1024.0f * 1024.0f))
					/ (time / 1000.0f)));
			float cfactor = done * 1.0f / doneEncoded;
			System.out.println("mean compression factor = " + cfactor);		
			System.out.println("files/sec = " + (runs * (args.length - 4) / (time / 1000.0f)));		
			System.out.println("checksum = " + checksum);			
		}
	}
	
	private static byte[] serializeWithXOM(Document doc) throws IOException {
//		return doc.toXML().getBytes();
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		Serializer ser = new Serializer(out);
		ser.write(doc);
		ser.flush();
		return out.toByteArray();
	}
	
	private static boolean equalsDocTypeEquals(DocType x, DocType y) {
		if (x != y) {
			if (x == null && y != null)
				return false;
			if (x != null && y == null)
				return false;
			
			if (!(x.getInternalDTDSubset().equals(y.getInternalDTDSubset())))
				return false;
			if (!(eq(x.getPublicID(), y.getPublicID())))
				return false;
			if (!(eq(x.getSystemID(), y.getSystemID())))
				return false;
			if (!(eq(x.getRootElementName(), y.getRootElementName())))
				return false;
			return true;
		}

		return true;
	}
	
	private static boolean eq(Object x, Object y) {
		if (x != y) {
			if (x == null && y != null)
				return false;
			if (x != null && y == null)
				return false;
			if (!x.equals(y))
				return false;
		}

		return true;
	}
	
	// ignore some stuff from the test suite dirs
	private static boolean ignore(String file) {
		if (file.endsWith(".html")) return true;
		if (file.endsWith(".out")) return true;
		if (file.endsWith(".txt")) return true;
		if (file.endsWith(".bat")) return true;
		if (file.endsWith(".sh")) return true;
		if (file.endsWith(".sxx")) return true;
		if (file.endsWith(".ssx")) return true;
		if (file.endsWith(".dtd")) return true;
		if (file.endsWith(".xq")) return true;
		if (file.endsWith(".zip")) return true;
		if (file.endsWith(".gz")) return true;
		if (file.endsWith(".jpg")) return true;
		if (file.endsWith(".gif")) return true;
//		if (file.endsWith("atrs13.xml")) return true; // char encoding
//		if (file.endsWith("atrs15.xml")) return true; // char encoding
		
//		if (file.endsWith("atrs47.xsl")) return true;
//		if (file.endsWith("atrs48.xsl")) return true;
//		if (file.endsWith("idky04.xml")) return true;
//		if (file.endsWith("idky10.xml")) return true;
//		if (file.endsWith("idky104.xml")) return true;
//		if (file.endsWith("idky11.xml")) return true;
//		if (file.endsWith("idky122.xml")) return true;
//		if (file.endsWith("idky123.xml")) return true;
		
		if (file.endsWith("nspc05x.xml")) return true; // not wellformed
//		if (file.endsWith("idky28.xml")) return true; // char encoding
//		if (file.endsWith("idky28wolf.xml")) return true; // char encoding
//		if (file.endsWith("idky35.xml")) return true; // char encoding
//		if (file.endsWith("stdxmlfile.xml")) return true; // char encoding
//		if (file.endsWith("copy19.xsl")) return true; // char encoding
//		if (file.endsWith("copy20.xsl")) return true; // char encoding
//		if (file.endsWith("copy22.xml")) return true; // char encoding
//		if (file.endsWith("copy38.xsl")) return true; // char encoding
		if (file.endsWith("ent22.xml")) return true; // not wellformed
		if (file.endsWith("2")) return true; // not wellformed

//		if (file.endsWith("expr85.xml")) return true; // char encoding
//		if (file.endsWith("numberformat06.xsl")) return true; // char encoding
//		if (file.endsWith("numberformat07.xsl")) return true; // char encoding
//		if (file.endsWith("str122.xsl")) return true; // char encoding
//		if (file.endsWith("str126.xsl")) return true; // char encoding
//		if (file.endsWith("str127.xsl")) return true; // char encoding
//		
//		if (file.indexOf("entref") >= 0) return true; // char encoding
		
		if (file.indexOf("err.") >= 0) return true;
		if (file.indexOf("ERR.") >= 0) return true;
//		if (file.indexOf("fail") >= 0) return true;
		if (file.indexOf("Untitled") >= 0) return true;
		
		// xom test suite:
		if (file.endsWith("CVS")) return true; // CVS messes up char encoding
		if (file.endsWith("ucs4bigendian.xml")) return true; // CVS messes up char encoding
		if (file.endsWith("ucs4littleendian.xml")) return true; // CVS messes up char encoding
		if (file.endsWith("utf8.xml")) return true; // CVS messes up char encoding
		if (file.endsWith("097.ent")) return true; // CVS messes up char encoding
		if (file.endsWith("test_ebcdic.xml")) return true; // CVS messes up char encoding
		
		// W3C XML test suite:
		if (file.endsWith(".ent")) return true; 
		if (file.endsWith("xmltest/valid/sa/012.xml")) return true; // invalid
		if (file.indexOf("fail") >= 0) return true;
		if (file.endsWith("p04pass1.xml")) return true; 
		if (file.endsWith("namespaces/1.0/009.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/010.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/011.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/013.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/014.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/015.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/016.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/023.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/025.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/026.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/029.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/030.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.1/005.xml")) return true; // not wellformed
		
//		if (file.endsWith("valid/sa/066.xml")) return true; // xerces-2.6.2 internal subset
//		if (file.endsWith("valid/sa/101.xml")) return true; // xerces-2.6.2 internal subset
//		if (file.endsWith("valid/sa/090.xml")) return true; // crimson internal subset
		
		return false;
	}
		
}