/*
 * Copyright (c) 2003, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.tests;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;

import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.NodeFactory;
import nu.xom.ParsingException;
import nux.xom.pool.BuilderFactory;
import nux.xom.pool.BuilderPool;
import nux.xom.pool.XOMUtil;

//import org.apache.log4j.Level;
//import org.apache.log4j.Logger;

/**
 * Command line program that validates a set of local XML document files against a W3C XML Schema.
 * Run this program without any arguments to get help (including descriptions of all options).
 * <p>
 * Requirements:
 * 1) JDK 1.4 or higher, 
 * 2) xerces-java @ http://xml.apache.org
 * 3) gnu-getopt @ http://www.urbanophile.com/%7Earenn/hacking/download.html 
 *
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.4 $, $Date: 2004/12/22 17:19:32 $
 */
public final class ValidationCommand {

	private LongOpt[] longopts;
	private StringBuffer sb;
	 	
	private int iterations;
	private String[] fileNames;
	private String namespace;
	private String schema;
	private boolean validate;
	private int maxPoolEntries;

//	private static final org.apache.commons.logging.Log log = org.apache.commons.logging.LogFactory.getLog(Validator.class);

	/**
	 * Run this program without any arguments to get help (including descriptions of all options).
	 */
	public static void main(String[] args) {
		new ValidationCommand().doMain(args);
	}
	
	private ValidationCommand() {
		initOptions();
	}
	
	private void doMain(String[] args) {
		try {
			if (parseArguments(args) < 0) return;
			run();
		} catch (UsageException e) {
			System.err.println(getProgramName() + ": " + e.getMessage());
			System.err.println(getShortUsage());
		} catch (Throwable t) {
			System.err.print("Oopsla: ");
			t.printStackTrace(System.err);
		}	
	}
	
	/**
	 * Define command line options, default value is defined on the same line.
	 */
	private void initOptions() {
		this.sb = new StringBuffer();
		ArrayList options = new ArrayList();
		options.add( new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h') );
		options.add( new LongOpt("version", LongOpt.NO_ARGUMENT, null, 'v') );		
		options.add( new LongOpt("wellformed", LongOpt.NO_ARGUMENT, null, 'V') ); this.validate = true;
		options.add( new LongOpt("namespace", LongOpt.REQUIRED_ARGUMENT, sb, 's') ); this.namespace = null;
		options.add( new LongOpt("schema", LongOpt.REQUIRED_ARGUMENT, sb, 'w') ); this.schema = null;
//		options.add( new LongOpt("loglevel", LongOpt.REQUIRED_ARGUMENT, sb, 'l') ); setLogLevels(Level.INFO);
		options.add( new LongOpt("nopool", LongOpt.NO_ARGUMENT, null, 'p') ); maxPoolEntries = 1;
		options.add( new LongOpt("iterations", LongOpt.REQUIRED_ARGUMENT, sb, 'i') ); this.iterations = 1;
			
		this.longopts = new LongOpt[options.size()];
		options.toArray(this.longopts);		
	}
	
	private int parseArguments(String[] args) {
		if (args.length == 0 ) {
			System.out.println(getLongUsage());
			return -1;
		}
		
		Getopt g = new Getopt(getProgramName(), args, ":", longopts, true);
		//Getopt g = new Getopt(getCommandString(), argv, "-:h:vf:b:pc:u:t:y:w:drq", longopts, true);
		//g.setOpterr(false); // We'll do our own error handling
	
		int c;
		while ((c = g.getopt()) != -1) {
//			log.trace("longind="+g.getLongind());
			switch (c) {
				case 'h' : // --help
					System.out.println(getLongUsage());
					return -1;
				case 'v' : // --version
					System.out.println(getVersionInfo());
					return -1;
				case 'V' :  // --wellformed
					this.validate = false;
					break;	
				case 'p' :  // --nopool
					maxPoolEntries = 0;
					break;	
				case 0 :  
					String arg = g.getOptarg();
					char val = (char) (new Integer(sb.toString())).intValue();
					String optionName = longopts[g.getLongind()].getName();
//					log.trace("Got long option with value '" + val + "' with argument " + ((arg != null) ? arg : "null"));
					switch (val) {
//						case 'l' : // --loglevel
//							setLogLevels(toLevel(arg));
//							break;
						case 's' :  // --namespace
							this.namespace = arg;
							break;	
						case 'w' : { // --schema
							// if the schema file location is a relative path, 
							// xerces interprets it relative to the XML instance document file, 
							// not the current working directory.
							// This may be surprising and errorprone, so we convert it to an absolute path 
							// Also, there are some obscure work arounds to make this work both on Unix and Windows, in all cases...
							this.schema = arg;
//							log.debug("schema="+schema);
							schema = getAbsoluteURI(schema);
//							log.debug("schema2="+schema);
							break;
						}
						case 'i' : // --iterations
							iterations = parseIntGreaterThanZero(arg, optionName);
							break;

						default :
							throw new InternalError("Oops. Should never reach here. val='" + val + "'");
					}				
					break;
				case ':' :
					throw new UsageException("Option '" + longopts[g.getLongind()].getName() + "' does indeed require an argument");
					//throw new UsageException("Argument missing for option " + (char) g.getOptopt() + ", errname=" + longopts[g.getLongind()].getName());
				case '?' :
					throw new UsageException("The option '" + (char) g.getOptopt() + "' is not valid");
				default :
					throw new InternalError("Oops. Should never reach here. getopt() returned '" + (char) c + "'");
			}
		}
	
		this.fileNames = parseNonOptionArguments(args, g.getOptind(), true, 1, Integer.MAX_VALUE);
			
		return 0;
	}
		
	/**
	 * Execute the query.
	 */	
	private void run() throws Exception {
//		log.trace("cwd=" + new File(".").getCanonicalPath());
//		log.trace("namespace=" + namespace);
//		log.trace("schema=" + schema);
		HashMap map = new HashMap();
		map.put(schema, namespace);
		
		BuilderPool pool = new BuilderPool(maxPoolEntries, 
			new BuilderFactory() {
				protected NodeFactory createNodeFactory() {
					return XOMUtil.getNullNodeFactory();
				}
			}
		);
		
		long start = System.currentTimeMillis();	
		
		for (int j=0; j < fileNames.length; j++) {
			for (int i=0; i < iterations; i++) { // do multiple times for benchmarking
				Builder builder = validate ? pool.getW3CBuilder(map) : pool.getBuilder(false);					
				try {				
					Document doc = builder.build(new File(this.fileNames[j]));
					//System.out.println(fileNames[j]);
					//if (log.isDebugEnabled()) log.debug(doc.toXML());
				} catch (ParsingException e) {
					String msg = validate ? "not valid" : "not well-formed";
					throw new UsageException(this.fileNames[j] + " is " + msg + ": " + e.getMessage() + " line: " + e.getLineNumber() + ", column: " + e.getColumnNumber(), e);
				}
				String msg = validate ? "valid and well-formed" : "well-formed";
				//if (log.isDebugEnabled()) log.debug(this.fileNames[j] + ": " + msg + ".");
				//System.out.println(this.fileNames[j] + ": " + msg + ".");
			}				
						
		}
		
		long end = System.currentTimeMillis();
		System.out.println("\n" + fileNames.length * iterations+ " files took " + (end-start)/1000.0f + " secs");
		System.out.println("--> " + (fileNames.length * iterations / ((end-start)/1000.0f)) + " files/secs");
	}
	
//	private void setLogLevels(Level level) {
//		Logger.getLogger(this.getClass()).setLevel(level);
//		Logger.getLogger(FireCmdLineProgram.class).setLevel(level);
//		Logger.getLogger(BuilderUtil.class).setLevel(level);
//	}
	
	private String getOptions() {
		return "[OPTION]... [FILE]...";
	}
	
	private String getLongUsage() {
		String namespace = "http://dsd.lbl.gov/p2pio-1.0";
		String text = 
	
		"\n" + getProgramName() + "-" + getVersion() + " - Validates a set of local XML files against an XML Schema.\n\n" +
	
		"Usage: " + getProgramName() + " " + getOptions() + "\n\n"+
	
		"Mandatory arguments to long options are mandatory for short options too.\n\n"+
	
		"Startup:\n"+
		"  -v,  --version             display the version of this program and exit.\n"+
		"       --help                print this help and exit.\n\n"+
		
		"Validation options:\n"+
		"  -w,  --wellformed          do not validate; check for well-formedness only.\n"+	
		"  -n,  --namespace=URI       namespace of external schema location.\n"+	
		"  -s,  --schema=FILE|URL     W3C XML schema of external schema location.\n"+	
		"       --nopool              disable pooling of XOM Builders.\n\n"+	
	
//		"Logging:\n"+
//		"  -l,  --loglevel=all|trace|debug|info|warn|error|fatal|off (default='info').\n\n"+
//	
		"Benchmarking:\n"+
		"  -i,  --iterations=INTEGER  execute query <iter> times (default=1).\n\n"+
	
		"Examples:\n"+
		"  " + "cd firefish/src/schemas\n"+
		"  " + getProgramName() + " --wellformed examples/ok.xml\n"+
		"  " + getProgramName() + " examples/ok.xml\n"+
		"  " + getProgramName() + " examples/ok2.xml\n"+
		"  " + getProgramName() + " --namespace=" + namespace + " --schema=p2pio.xsd examples/*.xml\n"+
		"  " + getProgramName() + " --namespace=" + namespace + " --schema=examples/../p2pio.xsd examples/*.xml\n"+
		"  " + getProgramName() + " --namespace=" + namespace + " --schema=http://dsd.lbl.gov/firefish/schemas/p2pio.xsd.xml examples/*.xml\n"+
		"  " + getProgramName() + " --schema=p2pio.xsd examples/*.xml\n";
	
		return text;
	}

	// hack to convert relative schema file names properly to an absolute URI,
	// taking care of Unix/Windows crossplatform issues
	private String getAbsoluteURI(String fileOrURI) {
		URI uri = null;
		try {
			uri = new URI(fileOrURI);
//			log.debug("uri="+uri);
//			log.debug("uri.getPath="+uri.getPath());
		}
		catch (URISyntaxException e) {
//			log.debug("URISyntaxException", e);
			uri = new File(fileOrURI).toURI();
		}
		
//		log.debug("uriScheme="+ uri.getScheme());
		if (uri.getScheme() == null) {
			uri = new File(fileOrURI).toURI();
//			log.debug("uriFile="+ uri);
		}

		if (uri.getScheme().equals("file")) {
			if (! new File(uri.getPath()).exists()) throw new UsageException("File '"+fileOrURI+"' not found.");
		}
		
		return uri.toString();
	}

	private String getMailAddress()   { return "Wolfgang Hoschek <whoschek@lbl.gov>"; }
	private String getProgramName()   { return "fire-validate"; }
	private String getVersion() { return "0.6.0"; }
	
	private String getShortUsage() {
		return 	
		"\nUsage: " + getProgramName() + " " + getOptions() +"\n\n"+
		"Try `" + getProgramName() + " --help' for more options.";
	}
	
	private String getVersionInfo() {
		return	
		"\n" + getProgramName() + "-" + getVersion() + " [Java " + System.getProperty("java.version") +"]" + "\n\n" +
		
		"Originally written by " + getMailAddress() + ".";
	}
	
	/** 
	 * Utility parse method for illegal/invalid option checking.
	 */ 
	private int parseIntGreaterThanZero(String arg, String optionName) {
		int val; 
		try {
			val = Integer.parseInt(arg);
			if (val <= 0) throw new NumberFormatException();
		}
		catch (NumberFormatException exc) {
			throw new UsageException("Invalid option '"+optionName+"="+arg+"'. Argument must be integer > 0.", exc);
		}
		return val;
	}
	
	/**
	 * Parse non-option command line arguments (option arguments have already been parsed).
	 */	
	private String[] parseNonOptionArguments(String[] args, int getOptind, boolean checkFile, int minArgs, int maxArgs) {
		if (minArgs < 0) throw new IllegalArgumentException("minArgs must be >= 0");
		if (maxArgs < 0) maxArgs = Integer.MAX_VALUE;
		
		for (int i = getOptind; i < args.length; i++) {
			//log.debug("Non-option argv element: " + args[i] + "\n");
		}
		if (getOptind > args.length-minArgs) {
			throw new UsageException("Further non-option arguments required but missing.");
		}
		if (getOptind < args.length-maxArgs) {
			throw new UsageException("Too many non-option arguments given; at most " + maxArgs + " allowed.");
		}
			
		String[] nonOptionArgs = new String[args.length - getOptind];
		for (int i = getOptind, j=0; i < args.length; i++, j++) {
			String nonOptionArg = args[i];
			nonOptionArgs[j] = nonOptionArg;
			if (checkFile && (! nonOptionArg.equals("-")) && (! new File(nonOptionArg).exists())) {
				throw new UsageException("File '"+nonOptionArg+"' not found.");
			}
		}
		return nonOptionArgs;
	}
	
	/**
	 * An exception used when a user defines invalid or illegal command line arguments.
	 */
	private static class UsageException extends RuntimeException {

		public UsageException() {
			super();
		} 
		
		public UsageException(String s) {
			super(s);
		}
		
		public UsageException(String message, Throwable cause) {
			super(message, cause);
		}
		
		public UsageException(Throwable cause) {
			super(cause);
		}
	}
	
}
