/*
 * Copyright (c) 2003, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.xquery;

import java.util.ArrayList;

import javax.xml.transform.TransformerException;

import net.sf.saxon.om.Axis;
import net.sf.saxon.om.AxisIterator;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.VirtualNode;
import net.sf.saxon.type.Type;
import net.sf.saxon.value.AnyURIValue;
import net.sf.saxon.value.AtomicValue;
import nu.xom.Attribute;
import nu.xom.Comment;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.Nodes;
import nu.xom.ProcessingInstruction;
import nu.xom.Text;

/**
 * A forward-only iterator representing an XQuery result sequence; allows to collect
 * execution output in streaming or batched manner.
 * <p>
 * Warning: Only to be used by XQuery.java
 * 
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.1 $, $Date: 2005/02/12 21:29:15 $
 */
final class DefaultResultSequence implements ResultSequence {
	
	/** the underlying saxon iterator producing result items */
	private SequenceIterator results;
	
	/** enable Nux debug output on System.err? */
	private static final boolean DEBUG = getSystemProperty("nux.xom.xquery.XQuery.debug", false);
	
	private static final Element ATOMIC_VALUE;
	static { // avoid reverification; also let's share strings via a template element
		ATOMIC_VALUE = new Element("atomic-value", "http://dsd.lbl.gov/nux");
		ATOMIC_VALUE.addAttribute(
			new Attribute("xsi:type", "http://www.w3.org/2001/XMLSchema-instance", ""));
	}
	
//	/** avoids reverification of qname and URI, as well as indexOf() calls, and saves string memory */
//	private final ReadElementMicroCache readElementCache = new ReadElementMicroCache(); 
//
//	/** avoids reverification of qname and URI, as well as indexOf() calls, and saves string memory */
//	private final ReadAttributeMicroCache readAttributeCache = new ReadAttributeMicroCache(); 
	
	DefaultResultSequence(SequenceIterator results) {
		this.results = results;
	}
	
	public Nodes toNodes() throws XQueryException {
		Nodes nodes = new Nodes();
		Node node;
		while ((node = next()) != null) {
			nodes.append(node);
		}
		return nodes;
	}
	
	public Node next() throws XQueryException {
		try {
			if (this.results != null) {
				Item item = this.results.next();
				if (item != null) return convertItem(item);
				this.results = null; // iterator exhausted; help gc
			}
			return null; 
		} catch (TransformerException e) {
			throw new XQueryException(e);
		}
	}
	
	/** Converts a saxon execution result node to XOM. */
	private static Node convertItem(Item item) {
		Object value = item;
		while (value instanceof VirtualNode) { // unwrap NodeWrapper et al
			value = ((VirtualNode) value).getUnderlyingNode();
		}
		if (DEBUG) System.err.println(
			"value.getClass=" + (value == null ? "null" : value.getClass().getName()));
		
		if (value instanceof Node) // the most common case; in particular XPath
			return (Node)value;
		else if (value instanceof NodeInfo) // coming from XQuery node constructors
			return convertNodeInfo((NodeInfo)value);
//		else if (value instanceof org.w3c.dom.Node) // now obsolete and never called
//			return convertDOM((org.w3c.dom.Node)value);
		else if (!(value instanceof AtomicValue)) { // should never happen 
			throw new IllegalArgumentException(
				"Oops! Expected atomic value but found non-atomic type: " + 
				(value == null ? "null" : value.getClass().getName()));
		}
		else return convertAtomicValue((AtomicValue)value);
	}
	
	/** Converts a saxon result node coming from XQuery node constructors to XOM. */
	private static Node convertNodeInfo(NodeInfo node) {
		if (DEBUG) System.err.println("convert nodeinfo: " + node.getClass().getName());
		switch (node.getNodeKind()) {
			case Type.ATTRIBUTE:
				return convertAttributeNodeInfo(node);
			case Type.ELEMENT: 
				return convertElementNodeInfo(node);
			case Type.DOCUMENT:
				return convertDocumentNodeInfo(node);
			case Type.TEXT:
				return new Text(node.getStringValue());
			case Type.COMMENT:
				return new Comment(node.getStringValue());
			case Type.PROCESSING_INSTRUCTION:
				return new ProcessingInstruction(node.getLocalPart(), node.getStringValue());
			case Type.NAMESPACE:
				return convertAtomicValue(new AnyURIValue(node.getStringValue()));
//				return new nu.xom.Namespace(node.getLocalPart(), node.getStringValue(), node.getParent());
			default: 
				throw new IllegalArgumentException(
					"Illegal NodeInfo kind: " + node.getClass().getName());
		}
	}
		
	private static Attribute convertAttributeNodeInfo(NodeInfo node) {
		// TODO: a goodie would be to try to infer an appropriate XOM Attribute.TYPE 
		// from NodeInfo.getTypeCode (also see ReceivingContentHandler)
		// but the typeCode is almost never present anyway...
		// take interned attr from cache, if available, avoiding XML reverification
		return new Attribute(
			node.getDisplayName(), node.getURI(), node.getStringValue());			
//		return readAttributeCache.build(
//		node.getDisplayName(), node.getURI(), node.getStringValue(), Attribute.Type.UNDECLARED);
	}
	
	private static Node convertDocumentNodeInfo(NodeInfo node) {
		Document doc = new Document(new Element("fakeRoot"));
		doc.setBaseURI(node.getBaseURI());
		
		boolean hasRootElement = false;
		int i = 0;
		NodeInfo next;
		AxisIterator iter = node.iterateAxis(Axis.CHILD);
		while ((next = (NodeInfo) iter.next()) != null) {
			Node child = convertNodeInfo(next);
			if (child instanceof Element) { // replace fake root with real root
				doc.setRootElement((Element) child); 
				hasRootElement = true;
			} else {
				doc.insertChild(child, i);
			}
			i++;
		}
		if (!hasRootElement) throw new IllegalArgumentException("Missing root element");
		return doc;			
	}
	
	private static Node convertElementNodeInfo(NodeInfo node) {
		if (DEBUG) System.err.println("converting element=" + node.getDisplayName());
		Element elem = new Element(node.getDisplayName(), node.getURI());
//		Element elem = this.readElementCache.build(node.getDisplayName(), node.getURI());
		NodeInfo next;
		
		// Append attributes
		AxisIterator iter = node.iterateAxis(Axis.ATTRIBUTE);
		ArrayList prefixes = null;
		while ((next = (NodeInfo) iter.next()) != null) {
			elem.addAttribute(convertAttributeNodeInfo(next));
			
			// keep track of attributes with prefixes, so we can avoid adding costly
			// additional namespaces declarations below. This safes potentially vast
			// amounts of memory due too XOM's expensive NS declaration storage scheme.
			String prefix = next.getPrefix();
			if (prefix.length() != 0) { // e.g. SOAP
				 if (prefixes == null) prefixes = new ArrayList(1);
				 prefixes.add(prefix);
			}
		}
		
		// Append namespace declarations (avoids unnecessary redeclarations).
		// For background see net.sf.saxon.om.NamespaceIterator and
		// net.sf.saxon.om.NamespaceDeclarationsImpl
		int[] namespaces = node.getDeclaredNamespaces(NodeInfo.EMPTY_NAMESPACE_LIST);
		int i = 0;
		int nsCode;
		NamePool pool = null;
		while (i < namespaces.length && (nsCode = namespaces[i]) != -1) {
			short uriCode = (short) (nsCode & 0xffff);
//			short prefixCode = (short) (nscode >> 16);
			if (uriCode != 0) { // it is not an undeclaration
				if (pool == null) pool = node.getNamePool();
				String uri = pool.getURIFromURICode(uriCode);
				String prefix = pool.getPrefixFromNamespaceCode(nsCode);
				
//				String currentURI = elem.getNamespaceURI(prefix);
//				if (currentURI != null && currentURI.equals(uri)) {				
				if (prefixes != null && prefixes.contains(prefix)) {
					if (DEBUG) System.err.println(
						"Safely ignoring additional namespace declaration for prefix=" 
						+ prefix + ", uri=" + uri);
				}
				else {
					if (DEBUG) System.err.println(
						"Adding additional namespace declaration for prefix=" 
						+ prefix + ", uri=" + uri);
					elem.addNamespaceDeclaration(prefix, uri);
				}
			}
			i++;
		}
		prefixes = null; // help gc
		namespaces = null; // help gc
		
//		// Append namespace declarations (would introduce lots of redundant redeclarations)
//		iter = node.iterateAxis(Axis.NAMESPACE);
//		while ((next = (NodeInfo) iter.next()) != null) {
//			String prefix = next.getLocalPart();
//			String uri = next.getStringValue();
//			if (DEBUG) System.err.println("converted prefix=" + prefix + ", uri=" + uri);
//			elem.addNamespaceDeclaration(prefix, uri);
//		}
		
		// Append children (recursively)
		iter = node.iterateAxis(Axis.CHILD);
		while ((next = (NodeInfo) iter.next()) != null) {
			elem.appendChild(convertNodeInfo(next));
		}
		
		return elem;
	}
	
	/** Converts saxon's atomic value to XOM. */
	private static Node convertAtomicValue(AtomicValue value) { 
		if (DEBUG) System.err.println("atomicValue.getClass="+value.getClass().getName());
		Element elem = new Element(ATOMIC_VALUE);  // copy to avoid reverification
		elem.getAttribute(0).setValue(value.getItemType().toString()); // e.g. "xs:integer"
		elem.appendChild(value.getStringValue()); // e.g. "123"
		return elem;
	}
	
	/** little helper for safe reading of boolean system properties */
	private static boolean getSystemProperty(String key, boolean def) {
		try { 
			return "true".equals(System.getProperty(key, String.valueOf(def)));
		} catch (Throwable e) { // better safe than sorry (applets, security managers, etc.) ...
			return def; // we can live with that
		}		
	}
	
//	/** Converts a saxon result node to XOM (method obsolete since saxon-8.3). */
//	private static Node convertDOM(org.w3c.dom.Node value) {
//		if (DEBUG) System.err.println("DOM class: " + value.getClass().getName());
//		if (value instanceof org.w3c.dom.Text) {						
//			return DOMConverter.convert((org.w3c.dom.Text)value);
//		}
//		else if (value instanceof org.w3c.dom.Attr) {						
//			return DOMConverter.convert((org.w3c.dom.Attr)value);
//		}
//		else if (value instanceof org.w3c.dom.Element) {
//			return DOMConverter.convert((org.w3c.dom.Element)value);
//		}
//		else if (value instanceof org.w3c.dom.Document) {
//			return DOMConverter.convert((org.w3c.dom.Document)value);
//		}
//		else if (value instanceof org.w3c.dom.Comment) {
//			return DOMConverter.convert((org.w3c.dom.Comment)value);
//		}
//		else if (value instanceof org.w3c.dom.ProcessingInstruction) {
//			return DOMConverter.convert((org.w3c.dom.ProcessingInstruction)value);
//		}		
//		else if (value instanceof org.w3c.dom.DocumentType) {
//			return DOMConverter.convert((org.w3c.dom.DocumentType)value);
//		}
//		else throw new IllegalArgumentException(
//			"unknown DOM node type - should never happen: " + value.getClass().getName());
//	}
		
}
