/*
 * Copyright (c) 2003, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.xquery;

import nu.xom.Element;
import nu.xom.Node;
import nu.xom.Nodes;
import nu.xom.ParentNode;
import nu.xom.Text;
import nux.xom.pool.XQueryPool;

/**
 * Various utilities avoiding redundant code in several classes.
 * 
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.7 $, $Date: 2005/02/10 20:54:30 $
 */
public class XQueryUtil {
	
	private XQueryUtil() {} // not instantiable

	/**
	 * Recursively walks the given node subtree and merges runs of consecutive
	 * (adjacent) {@link Text} nodes (if present) into a single Text node
	 * containing their string concatenation; Empty Text nodes are removed. 
	 * <p>
	 * If present, CDATA nodes are treated as Text nodes.
	 * The semantics are the same as with the DOM method
	 * {@link org.w3c.dom.Node#normalize() org.w3c.dom.Node.normalize()}.
	 * <p>
	 * Note that documents built by a {@link nu.xom.Builder} with the default
	 * {@link nu.xom.NodeFactory} are guaranteed to never have adjacent or empty
	 * Text nodes. However, subsequent manual removal or insertion of nodes to
	 * the tree can cause Text nodes to become adjacent, and updates can cause
	 * Text nodes to become empty.
	 * <p>
	 * Text normalization is necessary to achieve strictly standards-compliant
	 * XPath and XQuery semantics if a query compares or extracts the value of
	 * individual Text nodes that (unfortunately) happen to be adjacent to
	 * other Text nodes. Luckily, such use cases are rare in practical
	 * real-world scenarios and thus a user hardly ever needs to call this method
	 * before passing a XOM tree into XQuery or XPath.
	 * <p>
	 * Example Usage:
	 * 
	 * <pre>
	 * Element foo = new Element("foo");
	 * foo.appendChild("");
	 * foo.appendChild("bar");
	 * foo.appendChild("");
	 * 
	 * Element elem = new Element("elem");
	 * elem.appendChild("");
	 * elem.appendChild(foo);
	 * elem.appendChild("hello");
	 * elem.appendChild("world");
	 * elem.appendChild(foo.copy());
	 * elem.appendChild("");
	 * 
	 * XQueryUtil.normalizeTexts(elem);
	 * 
	 * for (int i = 0; i &lt; elem.getChildCount(); i++) {
	 *     System.out.println(elem.getChild(i).toString());
	 *     if (elem.getChild(i) instanceof Element) {
	 *         Element nested = (Element) elem.getChild(i);
	 *         for (int j = 0; j &lt; nested.getChildCount(); j++) {
	 *             System.out.println("    " + nested.getChild(j).toString());
	 *         }
	 *     }  
	 * }
	 * </pre>
	 * 
	 * returns the following normalized output:
	 * 
	 * <pre>
	 * [nu.xom.Element: foo]
	 *     [nu.xom.Text: bar]
	 * [nu.xom.Text: helloworld]
	 * [nu.xom.Element: foo]
	 *     [nu.xom.Text: bar]	
	 * </pre>
	 * 
	 * @param node
	 *            the subtree to normalize
	 */
	public static void normalizeTexts(ParentNode node) {
		/*
		 * Rather efficient implementation. At least in comparison to
		 * nu.xom.samples.TextMerger, which is not only buggy but also a textbook
		 * example for algorithmic performance desaster, spending 99% of time
		 * doing completely unnecessary work even if there are no adjacent
		 * Texts, in addition to degradation to quadratic performance
		 * in the worst case (where all Texts are adjacent).
		 */ 
		for (int i=node.getChildCount(); --i >= 0; ) {
			Node child = node.getChild(i);
			if (child instanceof Element) { // recursively walk the tree
				normalizeTexts((Element)child);
			}
			else if (child instanceof Text) {
				// scan to beginning of adjacent run, if any
				int j = i;
				while (--i >= 0 && node.getChild(i) instanceof Text) ;
				
				i++;
				if (j == i) { // remove isolated empty Text node
					if (child.getValue().length() == 0) node.removeChild(i);
				} 
				else { // found more than one adjacent Text node; merge them
					// append forwards and remove backwards to minimize memory copies of list elems 
					int k = i;
					StringBuffer buf = new StringBuffer(node.getChild(k++).getValue());
					while (k <= j) buf.append(node.getChild(k++).getValue());
					k = j;
					while (k >= i) node.removeChild(k--);
					
					// replace run with compact merged Text node unless empty
					if (buf.length() > 0) node.insertChild(new Text(buf.substring(0)), i);
				}
			}
		}
	}
	
	/**
	 * Executes the given W3C XQuery or XPath against the given context node
	 * (subtree); convenience method. Equivalent to
	 * 
	 * <pre>
	 * return XQueryPool.GLOBAL_POOL.getXQuery(query, null).execute(contextNode).toNodes();
	 * </pre>
	 * 
	 * Example usage:
	 * <pre>
	 * // find the atom named 'Zinc' in the periodic table:
	 * Document doc = new Builder().build(new File("samples/data/periodic.xml"));
	 * Node result = XQueryUtil.xquery(doc, "/PERIODIC_TABLE/ATOM[NAME = 'Zinc']").get(0);
	 * System.out.println("result=" + result.toXML());
	 * </pre>
	 * 
	 * @param contextNode
	 *            the context node to execute the query against. The context
	 *            node is available to the query as the value of the query
	 *            expression ".". If this parameter is <code>null</code>, the
	 *            context node will be undefined.
	 * @param query
	 *            the XQuery or XPath string
	 * @return the nodes of the result sequence.
	 * @throws RuntimeException
	 *             if an XQueryException occurs (unchecked exception for
	 *             convenience)
	 * @see XQuery
	 * @see XQueryPool
	 */
	public static Nodes xquery(Node contextNode, String query) {
		try {
			return XQueryPool.GLOBAL_POOL.getXQuery(query, null).execute(contextNode).toNodes();
		} catch (XQueryException e) { // part of the "convenience"
			throw new RuntimeException(e);
		}
	}
	
}